-- =============================================================================
-- Copyright © 2017-2018 by Cryptographic Engineering Research Group (CERG),
-- ECE Department, George Mason University
-- Fairfax, VA, U.S.A.
-- Author: Farnoud Farahmand
-- =============================================================================
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.numeric_std.all;
library work;
use work.ntru_pkg.all;

entity bpgm_mgf is
    generic(
    	BETA			: integer := 11;
    	W 				: integer := 32;
        SHA_INFIFO_LOG2DEPTH  : integer := 4;
        SHA_OUTFIFO_LOG2DEPTH : integer := 4
    );
    port(
    	-- inputs
    	clk				: in  std_logic;
    	rst				: in  std_logic;
        --! DFU input bus
        --! pdi_data (m_cm) ram bus
        pdi_data        : in std_logic_vector(W-1 downto 0);
        pdi_addr        : out std_logic_vector(5 downto 0);
        pdi_rd          : out std_logic;
        --! b_cb ram bus
        b               : in std_logic_vector(W-1 downto 0);
        b_addr          : out std_logic_vector(2 downto 0);
        b_rd            : out std_logic;
        --! h ram bus
        htrunc          : in std_logic_vector(W-1 downto 0);
        h_addr          : out std_logic_vector(2 downto 0);
        h_rd            : out std_logic;

        --! MGF input
        R4              : in std_logic_vector(127 downto 0);
        cR4             : in std_logic_vector(127 downto 0);

        R_fe_valid      : in std_logic;
        R_fe_read       : out std_logic;
        --! BPGM output FIFO bus
        bpgm_fifo_dout  : out std_logic_vector(BETA-1 downto 0);
        bpgm_fifo_empty : out std_logic;
        bpgm_fifo_read  : in std_logic;
        bpgm_fifo_almost_full : out std_logic;
        --! MGF mask bus
        mask_data       : out std_logic_vector(2*2*W -1 downto 0);
        mask_valid      : out std_logic;
        mask_ready      : in std_logic;
        --! Control signals
        start           : in std_logic;
        chk_failed_bwc1 : out std_logic;
        chk_failed_bwc2 : out std_logic;
        done            : out std_logic;
        decrypt         : in std_logic;
        sel_param       : in std_logic;
        sel_bpgm_mgf    : in std_logic;
        msg_len         : in std_logic_vector(7 downto 0)
    );
end bpgm_mgf;

architecture arch of bpgm_mgf is
    --! SHA in FIFO
    signal infifo_full      : std_logic;
    signal infifo_write     : std_logic;
    signal infifo_din       : std_logic_vector(W-1 downto 0);
    signal sha_infifo_empty : std_logic;
    signal src_read         : std_logic;
    signal sha_infifo_dout  : std_logic_vector(W-1 downto 0);
    --! SHA Out FIFO
    signal sha_outfifo_full : std_logic;
    signal dst_write        : std_logic;
    signal sha_dout         : std_logic_vector(W-1 downto 0);
    signal sha_outfifo_empty: std_logic;
    signal sha_outfifo_read : std_logic;
    signal sha_outfifo_dout : std_logic_vector(W-1 downto 0);
    --! SHA2 control signals
    signal backup_ready     : std_logic;
    signal en_backup        : std_logic;
    signal sel_init_state   : std_logic;
    signal sel_en_size      : std_logic;
    signal size_ready       : std_logic;
    signal rst_sha2         : std_logic;
    --! should be added later
    signal sha_done         : std_logic;
    --! BWC1 control signals
    signal outfifo_almost_full : std_logic;
    signal bwc1_infifo_read  : std_logic;
    signal bwc1_infifo_empty : std_logic;
    signal done_bwc1         : std_logic;
    signal rst_bwc1          : std_logic;
    --! BWC2 control signals
    signal bwc2_infifo_read  : std_logic;
    signal bwc2_infifo_empty : std_logic;
    signal done_bwc2         : std_logic;
    signal rst_bwc2          : std_logic;
    --! DFU signals
    signal pdi_cm            : std_logic_vector(W-1 downto 0);
    signal b_cb              : std_logic_vector(W-1 downto 0);
    signal dfu_en_reg        : std_logic;
    signal dfu_o             : std_logic_vector(W-1 downto 0);
    signal dfu_sel_i         : std_logic_vector(1 downto 0);
    signal dfu_sel_o         : std_logic_vector(2 downto 0);
    signal dfu_sel_oid       : std_logic;
    signal dfu_en_backup     : std_logic;
    signal dfu_sel_backup    : std_logic;

    --! Controler signals
    type t_state is (S_RESET, S_BPGM, S_FIRSTCHUNK_HDR, S_DATA_TRANSFER_BPGM_ENC,
        S_DATA_TRANSFER_MGF, S_WAIT_DTE,
        S_b_TRANSFER_ENC, S_h_TRANSFER_ENC, S_Ci_TRANSFER_ENC, S_LASTCHUNK_HDR,
        S_DONE, S_MGF, S_DATA_TRANSFER_BPGM_DEC, S_LASTCHUNK_HDR_MGF,
        S_LASTCHUNK_DATA_MGF, S_LASTCHUNK_DATA_BRAM_MGF);
        --S_DATA_TRANSFER_MGF_DEC
    signal state, state_next  : t_state;

    --! Registers
    signal c_pramaddr_next, c_pramaddr_r     : std_logic_vector(5 downto 0);
    signal c_bhramaddr_next, c_bhramaddr_r   : std_logic_vector(3 downto 0);
    signal ci_next, ci_r                     : std_logic_vector(31 downto 0);
    signal c_read32_next, c_read32_r         : std_logic_vector(7 downto 0);
    signal dfu_sel_o_next, dfu_sel_o_r       : std_logic_vector(2 downto 0);
    signal num_backup32_next, num_backup32_r : std_logic_vector(7 downto 0);
    signal prev_state_next, prev_state_r     : std_logic_vector(1 downto 0);
    signal lastchunk_8size_next, lastchunk_8size_r : std_logic_vector(6 downto 0);
    signal backup_ramaddr_next, backup_ramaddr_r   : std_logic_vector(5 downto 0);
    signal backable_next, backable_r         : std_logic;
    signal ci_2blk_next, ci_2blk_r           : std_logic;
    signal return_to_btrans_next, return_to_btrans_r  : std_logic;
    signal sel_init_state_next, sel_init_state_r      : std_logic;
    signal sel_en_size_next, sel_en_size_r            : std_logic;
    signal dfu_add_blk_next, dfu_add_blk_r            : std_logic;
    signal backed_up_next, backed_up_r                : std_logic;
    signal done_next, done_r                          : std_logic;

    signal in_size           : std_logic_vector(8 downto 0);
    signal in_size_ci        : std_logic_vector(8 downto 0);
    signal in_size_ci_mi64   : std_logic_vector(8 downto 0);
    signal bh_size           : integer;
    signal num_hash_req_bwc1 : integer;
    signal num_hash_req_bwc2 : integer;

    signal MGF_32bits_FirstChunk  : integer;
    signal MGF_Full_32bits_LastChunk  : integer;
    signal MGF_bits_LastChunk     : integer;
    signal MGF_LastBLK_PISO_Shift : integer;
    signal MGF_Partialbits_Last32Chunk : integer;
    signal MGF_32bits_LastChunk   : integer;

    signal msg_len_7to2      : std_logic_vector(5 downto 0);
    signal msg_len_1to0      : std_logic_vector(1 downto 0);

    signal en_piso, ld_piso, piso_rot : std_logic;
    signal piso_in1, piso_in2 : std_logic_vector(127 downto 0);
    signal piso_out           : std_logic_vector(W-1 downto 0);

    --! Not signals
    signal rstn                : std_logic;
    signal sha_infifo_empty_n  : std_logic;
    signal sha_outfifo_full_n  : std_logic;

    --! RFe_backup BRAM
    signal RFe_b_en, RFe_b_we  : std_logic;
    signal RFe_b_addr          : std_logic_vector(3 downto 0);
    signal RFe_b_di, RFe_b_do  : std_logic_vector(31 downto 0);

begin

--! CONTROLLER
p_fsm: process(clk)
begin
    if rising_edge(clk) then
        if (rst = '1') then
            state <= S_RESET;
            rst_bwc1    <= '1';
        else
            rst_bwc1    <= '0';
            state       <= state_next;
            -- i_valid_r   <= i_valid;
        end if;
        c_pramaddr_r       <= c_pramaddr_next;
        c_bhramaddr_r      <= c_bhramaddr_next;
        ci_r               <= ci_next;
        backable_r         <= backable_next;
        ci_2blk_r          <= ci_2blk_next;
        c_read32_r         <= c_read32_next;
        dfu_sel_o_r        <= dfu_sel_o_next;
        return_to_btrans_r <= return_to_btrans_next;
        sel_init_state_r   <= sel_init_state_next;
        sel_en_size_r      <= sel_en_size_next;
        dfu_add_blk_r      <= dfu_add_blk_next;
        num_backup32_r     <= num_backup32_next;
        prev_state_r       <= prev_state_next;
        lastchunk_8size_r  <= lastchunk_8size_next;
        backed_up_r        <= backed_up_next;
        backup_ramaddr_r   <= backup_ramaddr_next;
        done_r             <= done_next;
    end if;
end process;

p_comb: process(state, start, sel_bpgm_mgf, in_size_ci, in_size, backable_r,
    infifo_full, decrypt, c_read32_r, num_backup32_r, backed_up_r,
    c_pramaddr_r, msg_len, backup_ready, size_ready, c_bhramaddr_r, bh_size,
    dfu_sel_o_r, done_bwc1, done_bwc2, prev_state_r, ci_r, num_hash_req_bwc1,
    sha_done, dfu_o, msg_len_1to0, msg_len_7to2, lastchunk_8size_r,
    backup_ramaddr_r, R_fe_valid, piso_out, MGF_32bits_FirstChunk,
    num_hash_req_bwc2, RFe_b_do, MGF_LastBLK_PISO_Shift, infifo_din)
begin
    --! Default values
    state_next            <= state;
    c_pramaddr_next       <= c_pramaddr_r;
    c_bhramaddr_next      <= c_bhramaddr_r;
    ci_next               <= ci_r;
    backable_next         <= backable_r;
    ci_2blk_next          <= ci_2blk_r;
    c_read32_next         <= c_read32_r;
    dfu_sel_o_next        <= dfu_sel_o_r;
    return_to_btrans_next <= return_to_btrans_r;
    sel_init_state_next   <= sel_init_state_r;
    sel_en_size_next      <= sel_en_size_r;
    dfu_add_blk_next      <= dfu_add_blk_r;
    num_backup32_next     <= num_backup32_r;
    prev_state_next       <= prev_state_r;
    lastchunk_8size_next  <= lastchunk_8size_r;
    backed_up_next        <= backed_up_r;
    backup_ramaddr_next   <= backup_ramaddr_r;
    done_next             <= done_r;

    infifo_write    <= '0';
    infifo_din      <= (others => '0');
    pdi_addr        <= c_pramaddr_r;
    pdi_rd          <= '1';
    b_addr          <= c_bhramaddr_r(2 downto 0);
    b_rd            <= '1';
    h_addr          <= c_bhramaddr_r(2 downto 0);
    h_rd            <= '1';
    dfu_sel_i       <= "00";
    dfu_sel_oid     <= '0';
    dfu_sel_o       <= "000";
    dfu_en_reg      <= '0';
    dfu_en_backup   <= '0';
    dfu_sel_backup  <= '0';

    en_backup       <= '0';
    sel_en_size     <= sel_en_size_r;
    sel_init_state  <= sel_init_state_r;
    rst_sha2        <= '0';

    --rst_bwc1        <= '0';
    rst_bwc2        <= '0';

    R_fe_read       <= '0';
    en_piso         <= '0';
    ld_piso         <= '0';
    piso_rot        <= '0';

    --done            <= '0';

    RFe_b_en        <= '0';
    RFe_b_we        <= '0';
    RFe_b_addr      <= (others => '0');
    RFe_b_di        <= (others => '0');

    case state is
        when S_RESET =>
            if (start = '1') then
                c_pramaddr_next       <= (others => '0');
                c_bhramaddr_next      <= (others => '0');
                ci_next	              <= (others => '0');
                dfu_sel_o_next        <= (others => '0');
                c_read32_next         <= (others => '0');
                num_backup32_next     <= (others => '0');
                prev_state_next       <= (others => '0');
                lastchunk_8size_next  <= (others => '0');
                backup_ramaddr_next   <= (others => '0');
                dfu_add_blk_next      <= '0';
                backable_next         <= '0';
                ci_2blk_next          <= '0';
                return_to_btrans_next <= '0';
                sel_init_state_next   <= '0';
                sel_en_size_next      <= '0';
                backed_up_next        <= '0';
                done_next             <= '0';

                -- may need modification
                -- rst_bwc1              <= '1';
                rst_bwc2        <= '1';

                rst_sha2        <= '1';

                if (sel_bpgm_mgf = '1') then
                    state_next   <= S_MGF;
                else
                    state_next   <= S_BPGM;
                end if;
            end if;

        When S_BPGM =>
            if (in_size_ci(5 downto 0) = 0) then
                if (in_size_ci(8 downto 6) = 1) then
                    backable_next        <= '0';
                else
                    backable_next        <= '1';
                    num_backup32_next    <= "00" & in_size_ci_mi64(7 downto 2);
                    lastchunk_8size_next <= "1000000";
                end if;
                ci_2blk_next   <= '0';
            elsif (in_size_ci(5 downto 0) /= 0) and
                (in_size_ci(8 downto 6) /= in_size(8 downto 6))then
                if (in_size_ci(8 downto 6) = 1) then
                    backable_next        <= '0';
                else
                    backable_next        <= '1';
                    num_backup32_next    <= "00" & in_size_ci_mi64(7 downto 2);
                    lastchunk_8size_next <=  std_logic_vector(to_unsigned((64 +
                        to_integer(unsigned(in_size_ci(1 downto 0)))), lastchunk_8size_next'length));
                end if;
                ci_2blk_next   <= '1';
            else
                if (in_size_ci(8 downto 6) = "000") then
                    backable_next    <= '0';
                else
                    backable_next        <= '1';
                    num_backup32_next    <= '0'&in_size_ci(8 downto 6)&"0000";
                    lastchunk_8size_next <= '0'&in_size_ci(5 downto 0);
                end if;
                ci_2blk_next    <= '0';
            end if;
            state_next     <= S_FIRSTCHUNK_HDR;


        when S_FIRSTCHUNK_HDR =>
            if (backable_r = '1') then
                --! conversion to bit size
                infifo_din   <= "000"&X"0000"&num_backup32_r&"00000";--! First header
            else
                infifo_din   <= X"80000"&in_size_ci&"000";--! First and last header
            end if;
            if (infifo_full = '0') then
                infifo_write <= '1';
                c_pramaddr_next <= c_pramaddr_r + 1;
                state_next      <= S_DATA_TRANSFER_BPGM_ENC;
            end if;

        --! pdi_ram is assumed to be BRAM
        when S_DATA_TRANSFER_BPGM_ENC =>
            infifo_din      <= dfu_o;
            if (c_read32_r = num_backup32_r) and (backable_r = '1') and (backed_up_r = '0')then
                prev_state_next <= "00";
                state_next  <= S_LASTCHUNK_HDR;
            elsif (infifo_full = '0') then
                dfu_en_reg      <= '1';
                infifo_write    <= '1';
                c_read32_next   <= c_read32_r + 1;
                c_pramaddr_next <= c_pramaddr_r + 1;
                if (c_pramaddr_r = 1)then
                    dfu_sel_oid     <= '1';
                    if (msg_len = 1) then
                        c_pramaddr_next <= (others => '0');
                        state_next  <= S_b_TRANSFER_ENC;
                    end if;
                elsif (c_pramaddr_r >= 2)then
                    if (msg_len <= 4) then
                        dfu_sel_i <= "01";
                        c_bhramaddr_next <= c_bhramaddr_r + 1;
                        case msg_len(2 downto 0) is
                            when "010"  => dfu_sel_o <= "010";
                            when "011"  => dfu_sel_o <= "001";
                            when others => dfu_sel_o <= "000";
                        end case;
                        c_pramaddr_next <= (others => '0');
                        state_next  <= S_b_TRANSFER_ENC;
                    elsif (msg_len > 4)and (msg_len_1to0 < 2) then
                        if (c_pramaddr_r = msg_len_7to2+1)then
                            if (msg_len(1 downto 0) = "00") then
                                dfu_sel_i       <= "01";
                                c_bhramaddr_next <= c_bhramaddr_r + 1;
                                dfu_sel_o_next  <= "000";
                            else
                                dfu_sel_o_next <= "101";
                            end if;
                            c_pramaddr_next <= (others => '0');
                            state_next      <= S_b_TRANSFER_ENC;
                        end if;
                    else
                        if (c_pramaddr_r = msg_len_7to2+2)then
                            if (msg_len(1 downto 0) = "10") then
                                dfu_sel_o      <= "010";
                                dfu_sel_o_next <= "100";
                            else
                                dfu_sel_o      <= "001";
                                dfu_sel_o_next <= "011";
                            end if;
                            dfu_sel_i       <= "01";
                            c_bhramaddr_next <= c_bhramaddr_r + 1;
                            c_pramaddr_next <= (others => '0');
                            state_next      <= S_b_TRANSFER_ENC;
                        end if;
                    end if;
                end if;
            else
                state_next     <= S_WAIT_DTE;
            end if;
            if (backable_r = '1') then
                if (backup_ready = '1') then
                    en_backup       <= '1';
                    sel_init_state_next <= '1';
                elsif (size_ready = '1') then
                    sel_en_size_next <= '1';
                end if;
            end if;

        when S_WAIT_DTE =>
            if (infifo_full = '0') then
                pdi_addr       <= c_pramaddr_r -1;
                state_next     <= S_DATA_TRANSFER_BPGM_ENC;
            end if;

        --! b_ram and h_ram are assumed to be distributed ram
        when S_b_TRANSFER_ENC =>
            -- prev_state_next <= "01";
            infifo_din      <= dfu_o;
            dfu_sel_i       <= "01";
            dfu_sel_o       <= dfu_sel_o_r;
            if (c_read32_r = num_backup32_r) and (backable_r = '1') and (backed_up_r = '0')then
                prev_state_next <= "01";
                state_next  <= S_LASTCHUNK_HDR;
            elsif (infifo_full = '0') then
                dfu_en_reg       <= '1';
                infifo_write     <= '1';
                c_read32_next    <= c_read32_r + 1;
                c_bhramaddr_next <= c_bhramaddr_r + 1;
                if (c_bhramaddr_r = bh_size-1) then
                    state_next       <= S_h_TRANSFER_ENC;
                    c_bhramaddr_next <= (others => '0');
                else
                    c_bhramaddr_next <= c_bhramaddr_r + 1;

                end if;
            end if;
            if (backable_r = '1') then
                if (backup_ready = '1') then
                    en_backup       <= '1';
                    sel_init_state_next <= '1';
                elsif (size_ready = '1') then
                    sel_en_size_next <= '1';
                end if;
            end if;

        --! b_ram and h_ram are assumed to be distributed ram
        when S_h_TRANSFER_ENC =>
            -- prev_state_next <= "10";
            infifo_din      <= dfu_o;
            dfu_sel_i       <= "10";
            dfu_sel_o       <= dfu_sel_o_r;
            if (c_read32_r = num_backup32_r) and (backable_r = '1') and (backed_up_r = '0')then
                prev_state_next <= "10";
                state_next  <= S_LASTCHUNK_HDR;
            elsif (infifo_full = '0') then
                dfu_en_reg       <= '1';
                infifo_write     <= '1';
                c_read32_next    <= c_read32_r + 1;
                c_bhramaddr_next <= c_bhramaddr_r + 1;
                if (c_bhramaddr_r = bh_size-1) then
                    if (c_read32_r = num_backup32_r-1) and (backable_r = '1') then
                        prev_state_next  <= "11";
                        state_next       <= S_LASTCHUNK_HDR;
                    else
                        state_next       <= S_Ci_TRANSFER_ENC;
                    end if;
                    c_bhramaddr_next <= (others => '0');
                else
                    c_bhramaddr_next <= c_bhramaddr_r + 1;
                end if;
            else
            end if;
            if (backable_r = '1') then
                if (backup_ready = '1') then
                    en_backup       <= '1';
                    sel_init_state_next <= '1';
                elsif (size_ready = '1') then
                    sel_en_size_next <= '1';
                end if;
            end if;

        when S_Ci_TRANSFER_ENC =>
            dfu_sel_i    <= "11";
            dfu_sel_o    <= dfu_sel_o_r;
            if (infifo_full = '0') then
                if (dfu_sel_o_r = "101") then
                    infifo_din    <= dfu_o;
                    dfu_en_reg    <= '1';
                    infifo_write  <= '1';
                    ci_next     <= ci_r + 1;
                    if (done_bwc1 = '1') or (ci_r=num_hash_req_bwc1-1) then
                        state_next  <= S_DONE;
                    elsif (backable_r = '1') then
                        state_next  <= S_LASTCHUNK_HDR;
                    else
                        state_next  <= S_FIRSTCHUNK_HDR;
                    end if;
                elsif (dfu_sel_o_r = "000") or (dfu_sel_o_r = "011") or (dfu_sel_o_r = "100") then
                    if (dfu_add_blk_r = '1') then
                        dfu_add_blk_next <= '0';
                        --! this part can be removed in the future
                        case dfu_sel_o_r is
                            when "000"  => infifo_din <= dfu_o(31 downto  8)&X"00";
                            when "011"  => infifo_din <= dfu_o(31 downto 16)&X"0000";
                            when others => infifo_din <= dfu_o(31 downto 24)&X"000000";
                        end case;
                        infifo_write  <= '1';
                        ci_next     <= ci_r + 1;
                        if (done_bwc1 = '1') or (ci_r=num_hash_req_bwc1-1) then
                            state_next  <= S_DONE;
                        elsif (backable_r = '1') then
                            state_next  <= S_LASTCHUNK_HDR;
                        else
                            state_next  <= S_FIRSTCHUNK_HDR;
                        end if;
                    else
                        dfu_add_blk_next <= '1';
                        infifo_din       <= dfu_o;
                        dfu_en_reg       <= '1';
                        infifo_write     <= '1';
                    end if;
                end if;
            end if;

            if (backable_r = '1') then
                if (backup_ready = '1') then
                    en_backup       <= '1';
                    sel_init_state_next <= '1';
                elsif (size_ready = '1') then
                    sel_en_size_next <= '1';
                end if;
            end if;

        when S_LASTCHUNK_HDR =>
            --! convertion to bit size
            infifo_din   <= "10"&X"00000"&lastchunk_8size_r&"000";
            if (infifo_full = '0') then
                infifo_write  <= '1';
                case prev_state_r is
                    when "00"   =>
                        if (backed_up_r = '1') then
                            dfu_sel_backup   <= '1';
                            dfu_en_reg       <= '1';
                            c_pramaddr_next  <= backup_ramaddr_r;
                        else
                            dfu_en_backup       <= '1';
                            backup_ramaddr_next <= c_pramaddr_r;
                            backed_up_next      <= '1';
                        end if;
                        state_next  <= S_DATA_TRANSFER_BPGM_ENC;
                    when "01"   =>
                        if (backed_up_r = '1') then
                            dfu_sel_backup      <= '1';
                            dfu_en_reg          <= '1';
                            c_bhramaddr_next    <= backup_ramaddr_r(3 downto 0);
                        else
                            dfu_en_backup       <= '1';
                            backup_ramaddr_next <= "00"&c_bhramaddr_r;
                            backed_up_next      <= '1';
                        end if;
                        state_next  <= S_b_TRANSFER_ENC;
                    when "10"   =>
                        if (backed_up_r = '1') then
                            dfu_sel_backup      <= '1';
                            dfu_en_reg          <= '1';
                            c_bhramaddr_next    <= backup_ramaddr_r(3 downto 0);
                        else
                            dfu_en_backup       <= '1';
                            backup_ramaddr_next <= "00"&c_bhramaddr_r;
                            backed_up_next      <= '1';
                        end if;
                        state_next  <= S_h_TRANSFER_ENC;
                    when others =>
                        if (backed_up_r = '1') then
                            dfu_sel_backup      <= '1';
                            dfu_en_reg          <= '1';
                        else
                            dfu_en_backup       <= '1';
                            backed_up_next      <= '1';
                        end if;
                        state_next  <= S_Ci_TRANSFER_ENC;
                end case;
            end if;

        When S_MGF =>
            backable_next    <= '1';
            infifo_din   <= std_logic_vector(to_unsigned(MGF_32bits_FirstChunk, 27)) &"00000";
            if (infifo_full = '0') and (R_fe_valid = '1')then
                infifo_write <= '1';
                R_fe_read    <= '1';
                en_piso      <= '1';
                ld_piso      <= '1';
                state_next <= S_DATA_TRANSFER_MGF;
            end if;

        When S_DATA_TRANSFER_MGF =>
            infifo_din    <= piso_out;
            if (infifo_full = '0') then
                if (c_read32_r(1 downto 0) = "11") then
                    if (R_fe_valid = '1') then
                        R_fe_read     <= '1';
                        ld_piso       <= '1';
                        en_piso       <= '1';
                        infifo_write  <= '1';
                        if (c_read32_r = MGF_32bits_FirstChunk-1) then
                            c_read32_next  <= (others =>'0');
                            state_next <= S_LASTCHUNK_HDR_MGF;
                        else
                            c_read32_next <= c_read32_r + 1;
                        end if;
                    end if;
                else
                    c_read32_next  <= c_read32_r + 1;
                    infifo_write  <= '1';
                    en_piso       <= '1';
                end if;
            end if;

            if (backup_ready = '1') then
                en_backup       <= '1';
                sel_init_state_next <= '1';
            elsif (size_ready = '1') then
                sel_en_size_next <= '1';
            end if;

        when S_LASTCHUNK_HDR_MGF =>
            if (infifo_full = '0') then
                infifo_write  <= '1';
                infifo_din <= '1'&std_logic_vector(to_unsigned(MGF_bits_LastChunk, 31));
                RFe_b_en   <= '1';
                RFe_b_we   <= '1';
                RFe_b_addr <= c_read32_r(3 downto 0);
                RFe_b_di   <= infifo_din;
                state_next <= S_LASTCHUNK_DATA_MGF;
            end if;

        when S_LASTCHUNK_DATA_MGF =>
            if (infifo_full = '0') then
                if (PARAM_N = 1087) then
                    infifo_write  <= '1';
                    RFe_b_en      <= '1';
                    if (c_read32_r > MGF_LastBLK_PISO_Shift) then
                        infifo_din    <= ci_r;
                        ci_next       <= ci_r + 1;
                        c_read32_next <= x"01";
                        RFe_b_addr    <= ZEROs(3 downto 0);
                        state_next    <= S_LASTCHUNK_DATA_BRAM_MGF;
                    else
                        RFe_b_we   <= '1';
                        RFe_b_addr <= (c_read32_r(3 downto 0))+1;
                        RFe_b_di   <= infifo_din;

                        infifo_din    <= piso_out;
                        en_piso       <= '1';
                        c_read32_next <= c_read32_r + 1;
                    end if;
                else
                    if (c_read32_r >= MGF_Full_32bits_LastChunk) then
                        RFe_b_en      <= '1';
                        if (c_read32_r(1 downto 0) > MGF_LastBLK_PISO_Shift) then
                            infifo_write  <= '1';
                            infifo_din  <= ci_r(MGF_Partialbits_Last32Chunk_1499-1 downto 0) &
                                    ZEROs(31 downto MGF_Partialbits_Last32Chunk_1499);
                            ci_next       <= ci_r + 1;
                            c_read32_next <= x"01";
                            RFe_b_addr    <= ZEROs(3 downto 0);
                            state_next    <= S_LASTCHUNK_DATA_BRAM_MGF;
                        elsif (c_read32_r(1 downto 0) = MGF_LastBLK_PISO_Shift) then
                            RFe_b_we   <= '1';
                            RFe_b_addr <= (c_read32_r(3 downto 0))+1;
                            RFe_b_di   <= infifo_din;
                            infifo_write  <= '1';
                            infifo_din  <= piso_out(31 downto 32-MGF_Partialbits_Last32Chunk_1499) &
                                    ci_r(31 downto MGF_Partialbits_Last32Chunk_1499);
                            en_piso       <= '1';
                            c_read32_next <= c_read32_r + 1;
                        else
                            RFe_b_we   <= '1';
                            RFe_b_addr <= (c_read32_r(3 downto 0))+1;
                            RFe_b_di   <= infifo_din;

                            infifo_write  <= '1';
                            infifo_din    <= piso_out;
                            en_piso       <= '1';
                            c_read32_next <= c_read32_r + 1;
                        end if;
                    else
                        infifo_din    <= piso_out;
                        RFe_b_di      <= infifo_din;
                        if (c_read32_r(1 downto 0) = "11") then
                            if (R_fe_valid = '1') then
                                R_fe_read     <= '1';
                                ld_piso       <= '1';
                                en_piso       <= '1';
                                infifo_write  <= '1';
                                RFe_b_en      <= '1';
                                RFe_b_we      <= '1';
                                RFe_b_addr    <= (c_read32_r(3 downto 0))+1;
                                c_read32_next <= c_read32_r + 1;
                            end if;
                        else
                            c_read32_next <= c_read32_r + 1;
                            infifo_write  <= '1';
                            RFe_b_en      <= '1';
                            RFe_b_we      <= '1';
                            RFe_b_addr    <= (c_read32_r(3 downto 0))+1;
                            en_piso       <= '1';
                        end if;
                    end if;
                end if;

                if (backup_ready = '1') then
                    en_backup       <= '1';
                    sel_init_state_next <= '1';
                elsif (size_ready = '1') then
                    sel_en_size_next <= '1';
                end if;

            end if;

        when S_LASTCHUNK_DATA_BRAM_MGF =>
            if (infifo_full = '0') then
                if (c_read32_r = MGF_32bits_LastChunk+1) then
                    if (PARAM_N = 1087) then
                        infifo_din    <= ci_r;
                    else
                        infifo_din    <= ci_r(MGF_Partialbits_Last32Chunk_1499-1 downto 0) &
                            ZEROs(31 downto MGF_Partialbits_Last32Chunk_1499);
                    end if;
                    infifo_write  <= '1';
                    RFe_b_en      <= '1';
                    RFe_b_addr    <= ZEROs(3 downto 0);
                    c_read32_next <= x"01";
                    ci_next       <= ci_r + 1;
                    if (done_bwc2 = '1') or (ci_r=num_hash_req_bwc2-1)then
                        state_next    <= S_DONE;
                    end if;
                else
                    infifo_din    <= RFe_b_do;
                    infifo_write  <= '1';
                    RFe_b_en      <= '1';
                    RFe_b_addr    <= c_read32_r(3 downto 0);
                    c_read32_next <= c_read32_r + 1;
                end if;
            end if;

            if (backup_ready = '1') then
                en_backup       <= '1';
                sel_init_state_next <= '1';
            elsif (size_ready = '1') then
                sel_en_size_next <= '1';
            end if;

        when S_DONE =>
            if (sel_bpgm_mgf = '0') then
                if (done_bwc1 = '1') then
                    done_next <= '1';
                    state_next  <= S_RESET;
                end if;
            else
                if (done_bwc2 = '1')then
                    done_next <= '1';
                    state_next  <= S_RESET;
                end if;
            end if;

        when S_DATA_TRANSFER_BPGM_DEC =>
            state_next  <= S_RESET;

    end case;
end process;

--! DATAPATH

done <= done_r;

--! We assume that msg_len and decrypt signals are registered externally in
--! top level controller
Param_data_size_1087: if PARAM_N = 1087 generate
    in_size <= std_logic_vector(to_unsigned((to_integer(unsigned(msg_len)) +
        Total_BPGM_Bytes_1087), in_size'length));
    bh_size  <= db_32_1087;
    num_hash_req_bwc1 <= NUM_HASH_REQ_BWC1_1087;
    num_hash_req_bwc2 <= NUM_HASH_REQ_BWC2_1087;

    MGF_32bits_FirstChunk  <= MGF_32bits_FirstChunk_1087;
    MGF_Full_32bits_LastChunk  <= MGF_Full_32bits_LastChunk_1087;
    MGF_bits_LastChunk     <= MGF_bits_LastChunk_1087;
    MGF_LastBLK_PISO_Shift <= MGF_LastBLK_PISO_Shift_1087;
    MGF_Partialbits_Last32Chunk <= MGF_Partialbits_Last32Chunk_1087;
    MGF_32bits_LastChunk   <= MGF_32bits_LastChunk_1087;
end generate;

Param_data_size_1499: if PARAM_N = 1499 generate
    in_size <= std_logic_vector(to_unsigned((to_integer(unsigned(msg_len)) +
        Total_BPGM_Bytes_1499), in_size'length));
    bh_size  <= db_32_1499;
    num_hash_req_bwc1 <= NUM_HASH_REQ_BWC1_1499;
    num_hash_req_bwc2 <= NUM_HASH_REQ_BWC2_1499;

    MGF_32bits_FirstChunk  <= MGF_32bits_FirstChunk_1499;
    MGF_Full_32bits_LastChunk  <= MGF_Full_32bits_LastChunk_1499;
    MGF_bits_LastChunk     <= MGF_bits_LastChunk_1499;
    MGF_LastBLK_PISO_Shift <= MGF_LastBLK_PISO_Shift_1499;
    MGF_Partialbits_Last32Chunk <= MGF_Partialbits_Last32Chunk_1499;
    MGF_32bits_LastChunk   <= MGF_32bits_LastChunk_1499;
end generate;
--! Input size plus 32-bit counter
in_size_ci <= in_size + Ci_Byte;
--! Input size plus 32-bit counter minus 512 (one block of SHA-2)
in_size_ci_mi64 <= in_size_ci - 64;

msg_len_1to0 <= msg_len(1 downto 0);
msg_len_7to2 <= msg_len(7 downto 2);

-- pdi_cm <= cm when (decrypt = '1') else pdi_data;
pdi_cm <= pdi_data;
-- b_cb   <= cb when (decrypt = '1') else b;
b_cb   <= b;

piso_in1 <= cR4 when (decrypt = '1') else R4;
-- piso_in2 <= ZEROS(95 downto 0)&piso_out when (piso_rot = '1') else piso_in1;

-- Not signals
rstn <= not rst;
sha_infifo_empty_n <= not sha_infifo_empty;
sha_outfifo_full_n <= not sha_outfifo_full;

DFU: entity work.dfu(arch)
    generic map(W => W)
    port map(
        clk       => clk,
        rst       => rst,
        a         => pdi_cm,
        b         => b_cb,
        c         => htrunc,
        d         => ci_r,
        en_reg    => dfu_en_reg,
        en_backup_reg => dfu_en_backup,
        sel_backup    => dfu_sel_backup,
        sel_i     => dfu_sel_i,
        sel_oid   => dfu_sel_oid,
        sel_o     => dfu_sel_o,
        r         => dfu_o
    );

Piso: entity work.piso
    generic map(
        N => 128,
        M => W
        )
    port map(
        clk    => clk,
        rst    => rst,
        en     => en_piso,
        ld     => ld_piso,
        input  => piso_in1,
        output => piso_out
    );

RFe_Backup: entity work.rams_sp_rf_rst(syn)
    generic map(
    W => 32,
    addr_W => 4,
    addr_S => 16
    )
    port map(
        clk   => clk,
        en    => RFe_b_en,
        we    => RFe_b_we,
        rst   => rst,
        addr  => RFe_b_addr,
        di    => RFe_b_di,
        do    => RFe_b_do
    );

SHA_Input_FIFO_FWFT: entity work.fifo(structure)
    generic map(
        G_LOG2DEPTH    => SHA_INFIFO_LOG2DEPTH,
        G_W            => W
    )
    port map (
        clk              => clk,
        rstn             => rstn,
        full             => infifo_full,
        write            => infifo_write,
        di_data          => infifo_din,
        empty            => sha_infifo_empty,
        read             => src_read,
        do_data          => sha_infifo_dout,
        almost_full      => open,
        almost_empty     => open
    );

modified_SHA2: entity work.sha2_top(structure)
    -- generic map( hs => hashsize )
    port map(
        rst 	=> rst_sha2, --rst
        clk 	=> clk,
        din			=> sha_infifo_dout,
        src_read	=> src_read,
        src_ready	=> sha_infifo_empty,
        dout		=> sha_dout,
        dst_write	=> dst_write,
        dst_ready	=> sha_outfifo_full,
        --! NTRU control signals
        en_backup    => en_backup,
        backup_ready => backup_ready,
        sel_init_state   => sel_init_state,
        size_ready   => size_ready,
        sel_en_size  => sel_en_size
    );

SHA_Output_FIFO_FWFT: entity work.fifo(structure)
    generic map(
        G_LOG2DEPTH    => SHA_OUTFIFO_LOG2DEPTH,
        G_W            => W
    )
    port map (
        clk              => clk,
        rstn             => rstn,
        full             => sha_outfifo_full,
        write            => dst_write,
        di_data          => sha_dout,
        empty            => sha_outfifo_empty,
        read             => sha_outfifo_read,
        do_data          => sha_outfifo_dout,
        almost_full      => open,
        almost_empty     => open
    );

--! FIFO Interconnect
sha_outfifo_read  <= bwc2_infifo_read when (sel_bpgm_mgf = '1') else bwc1_infifo_read;
bwc1_infifo_empty <= '1' when (sel_bpgm_mgf = '1') else sha_outfifo_empty;
bwc2_infifo_empty <= sha_outfifo_empty when (sel_bpgm_mgf = '1') else '1';

BWC1: entity work.bwc1(arch)
    generic map(
        BETA => BETA,
        W    => W,
        -- INFIFO_LOG2DEPTH  => 5,
        OUTFIFO_LOG2DEPTH => 5
    )
    port map(
        rst 	=> rst_bwc1,
        clk 	=> clk,
        --! Input FIFO ports
        infifo_dout  => sha_outfifo_dout,
        infifo_read  => bwc1_infifo_read,
        infifo_empty => bwc1_infifo_empty,
        --! Output FWFT FIFO ports
        fifo_dout  => bpgm_fifo_dout,
        fifo_empty_valid => bpgm_fifo_empty,
        fifo_read  => bpgm_fifo_read,
        chk_failed => chk_failed_bwc1,
        done       => done_bwc1
    );

BWC2: entity work.bwc2(arch)
    generic map(
        W    => W
        -- INFIFO_LOG2DEPTH  => 5,
        -- OUTFIFO_LOG2DEPTH => 2
    )
    port map(
        rst 	=> rst_bwc2,
        clk 	=> clk,
        done    => done_bwc2,
        chk_failed => chk_failed_bwc2,
        --! Input FIFO ports
        infifo_dout  => sha_outfifo_dout,
        infifo_read  => bwc2_infifo_read,
        infifo_empty => bwc2_infifo_empty,
        --! Output handshaking ports
        output_data  => mask_data,
        output_valid => mask_valid,
        output_ready => mask_ready
    );

end arch;
