-- =============================================================================
-- Copyright © 2017-2018 by Cryptographic Engineering Research Group (CERG),
-- ECE Department, George Mason University
-- Fairfax, VA, U.S.A.
-- Author: Farnoud Farahmand
-- =============================================================================
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_arith.all;
library work;
use work.ntru_pkg.all;

entity bwc1 is
    generic(
        BETA : integer := 11;
    	C    : integer := 13;
        W    : integer := 32;
        OUTFIFO_LOG2DEPTH : integer := 2
    );
    port(
        clk        : in std_logic;
        rst        : in std_logic;
        --! Input FIFO ports
        infifo_dout  : in std_logic_vector(W-1 downto 0);
        infifo_read  : out std_logic;
        infifo_empty : in std_logic;
        --! Output FWFT FIFO ports
        fifo_dout  : out std_logic_vector(BETA-1 downto 0);
        fifo_empty_valid : out std_logic;
        fifo_read  : in std_logic;
        --! control signals
        chk_failed : out std_logic;
        done       : out std_logic
    );
end bwc1;

architecture arch of bwc1 is
    --! constants
    constant piso_width : integer := 16; -- must be divisible by N (256)
    constant vs_width   : integer := 28; -- min number bits needed for c (13)
    constant shft_width : integer := 4;
    constant ram_data_width : integer := 1;
    constant ZERO : std_logic_vector(W-1 downto 0) := (others => '0');
    constant ONES : std_logic_vector(W-1 downto 0) := (others => '1');

    --! output fifo
	signal outfifo_empty : std_logic;
    signal outfifo_read  : std_logic;
    signal outfifo_write : std_logic;
    signal outfifo_full  : std_logic;

    signal piso_out       : std_logic_vector(piso_width-1 downto 0);
    signal piso_r         : std_logic_vector(piso_width-2 downto 0);
    signal vs_out         : std_logic_vector(vs_width-1 downto 0);
    signal vs_rom_out     : std_logic_vector(shft_width-1 downto 0);
    signal mod_1499       : std_logic_vector(BETA-1 downto 0);
    signal mod_1087       : std_logic_vector(BETA-1 downto 0);
    signal from_mod_mux, from_mod_mux_r : std_logic_vector(BETA-1 downto 0);
    signal from_lseq_mux, from_lseq_mux_r  : std_logic;
    signal lseq_7495      : std_logic;
    signal lseq_7609      : std_logic;
    signal dout_invalid   : std_logic;
    signal dout_invalid_stdv : std_logic_vector(ram_data_width-1 downto 0);
    signal ram_addr_a     : std_logic_vector(BETA-1 downto 0);

    --! Control signals
    type t_state is (S_RESET, S_INIT, S_FIRTS_LOAD, S_SHIFT_LOAD, S_WAIT);
    -- S_LOAD,
    signal state, state_next  : t_state;
    signal checkpoint, checkpoint_next : std_logic_vector(3 downto 0);
    signal count_next, count_r      : std_logic_vector(CNT_SIZE -1 downto 0);
    signal c_Rcoef_next, c_Rcoef_r  : std_logic_vector(7 downto 0);
    signal c_hashw_next, c_hashw_r  : std_logic_vector(7 downto 0);
    signal c_hashw32_next, c_hashw32_r: std_logic_vector(6 downto 0);
    signal c_Wcoef_next, c_Wcoef_r  : std_logic_vector(7 downto 0);
    signal sel_init_ram  : std_logic;
    signal we_zero       : std_logic;
    signal we_one        : std_logic;
    signal en_piso       : std_logic;
    signal ld_piso       : std_logic;
    signal i_valid       : std_logic;
    signal hold_we_i     : std_logic;
    signal first_coef_next, first_coef_r : std_logic;
    signal wait_next, wait_r   : std_logic;
    signal done_next, done_r   : std_logic;
    signal load_next, load_r   : std_logic;
    signal bypass_empty_next, bypass_empty_r : std_logic;
    signal chk_failed_next, chk_failed_r : std_logic;
    signal hash_done_next, hash_done_r   : std_logic;

    signal rstn             : std_logic;

    signal hashw_flag       : std_logic;
    signal Wcoef_flag       : std_logic;

    signal bypass_empty     : std_logic;

    signal NUM_HASH32_REQ   : integer;

begin

--! CONTROLLER
p_fsm: process(clk)
begin
    if rising_edge(clk) then
        if (rst = '1') then
            state      <= S_RESET;
            checkpoint <= "0000";
        else
            state            <= state_next;
            checkpoint       <= checkpoint_next;
            from_lseq_mux_r  <= from_lseq_mux;
        end if;
        count_r        <= count_next;
        first_coef_r   <= first_coef_next;
        wait_r         <= wait_next;
        bypass_empty_r <= bypass_empty_next;
        c_Rcoef_r      <= c_Rcoef_next;
        c_Wcoef_r      <= c_Wcoef_next;
        c_hashw_r      <= c_hashw_next;
        c_hashw32_r    <= c_hashw32_next;
        done_r         <= done_next;
        load_r         <= load_next;
        chk_failed_r   <= chk_failed_next;
        hash_done_r    <= hash_done_next;
    end if;
end process;

p_comb: process(state, checkpoint, count_r, infifo_empty, dout_invalid,
    outfifo_full, vs_rom_out, first_coef_r, c_Rcoef_r, c_Wcoef_r, c_hashw_r,
    load_r, chk_failed_r, fifo_read, c_hashw32_r, i_valid, done_r, wait_r,
    hash_done_r, NUM_HASH32_REQ, wcoef_flag, hashw_flag, from_lseq_mux)
begin
    --! Default values
    state_next      <= state;
    checkpoint_next <= checkpoint;
    count_next      <= count_r;
    first_coef_next <= first_coef_r;
    wait_next       <= wait_r;
    c_Rcoef_next    <= c_Rcoef_r;
    c_Wcoef_next    <= c_Wcoef_r;
    c_hashw_next    <= c_hashw_r;
    c_hashw32_next  <= c_hashw32_r;
    done_next       <= done_r;
    load_next       <= load_r;
    hash_done_next  <= hash_done_r;
    chk_failed_next <= chk_failed_r;
    infifo_read     <= '0';
    we_zero         <= '0';
    we_one          <= '0';
    en_piso         <= '0';
    ld_piso         <= '0';
    outfifo_write   <= '0';
    done            <= '0';
    bypass_empty    <= '1';

    case state is
        when S_RESET =>
            count_next	    <= (others => '0');
            c_Rcoef_next	<= (others => '0');
            c_Wcoef_next	<= (others => '0');
            c_hashw_next    <= (others => '0');
            c_hashw32_next  <= (others => '0');
            first_coef_next <= '1';
            wait_next       <= '0';
            done_next       <= '0';
            load_next       <= '0';
            hash_done_next  <= '0';
            chk_failed_next <= '0';
            state_next      <= S_INIT;

        when S_INIT =>
            we_zero      <= '1';
            if (count_r = ONES(CNT_SIZE-1 downto 0)) then
                count_next	<= (others => '0');
                state_next  <= S_FIRTS_LOAD;
            else
                count_next <= count_r + 1;
            end if;

        when S_FIRTS_LOAD =>
            if (infifo_empty = '0') then
                infifo_read <= '1';
                en_piso     <= '1';
                ld_piso     <= '1';
                c_hashw32_next  <= c_hashw32_r + 1;
                state_next  <= S_SHIFT_LOAD;
            end if;

        when S_SHIFT_LOAD =>

            if (done_r = '1') or (chk_failed_r = '1') then
                state_next  <= S_WAIT;
            else
                if (hash_done_r = '0') then
                    if (load_r = '0') then
                        if (outfifo_full = '0') then
                            if (first_coef_r = '1') then
                                first_coef_next <= '0';
                                c_hashw_next    <= c_hashw_r + 1;
                                we_one    <= from_lseq_mux;
                            else
                                c_hashw_next    <= c_hashw_r + 1;
                                we_one    <= from_lseq_mux;
                                if (i_valid = '1') then
                                    c_Wcoef_next   <= c_Wcoef_r+1;
                                    outfifo_write <= '1';
                                end if;
                            end if;
                            if (vs_rom_out < 13)then
                                en_piso     <= '1';
                                load_next   <= '1';
                            end if;

                            if (count_r < 15) then
                                count_next    <= count_r + 1;
                            else
                                count_next	<= (others => '0');
                            end if;
                        end if;

                    else

                        if (outfifo_full = '0') then
                            if (c_hashw32_r = NUM_HASH32_REQ) then
                                c_hashw_next    <= c_hashw_r + 1;
                                we_one    <= from_lseq_mux;
                                if (i_valid = '1') then
                                    c_Wcoef_next   <= c_Wcoef_r+1;
                                    outfifo_write <= '1';
                                end if;

                                if (vs_rom_out < 13)then
                                    hash_done_next <= '1';
                                end if;

                                if (count_r < 15) then
                                    count_next    <= count_r + 1;
                                else
                                    count_next	<= (others => '0');
                                end if;

                            else
                                if (infifo_empty = '0') then
                                    if (wait_r = '0') then
                                        c_hashw_next    <= c_hashw_r + 1;
                                        we_one    <= from_lseq_mux;
                                        if (i_valid = '1') then
                                            c_Wcoef_next   <= c_Wcoef_r+1;
                                            outfifo_write <= '1';
                                        end if;
                                    else
                                        wait_next <= '0';
                                    end if;
                                    if (vs_rom_out < 13)then
                                        infifo_read <= '1';
                                        en_piso     <= '1';
                                        ld_piso     <= '1';
                                        load_next   <= '0';
                                        c_hashw32_next    <= c_hashw32_r + 1;
                                    end if;
                                    if (count_r < 15) then
                                        count_next    <= count_r + 1;
                                    else
                                        count_next	<= (others => '0');
                                    end if;
                                else
                                    if (wait_r = '0') then
                                        c_hashw_next    <= c_hashw_r + 1;
                                        we_one    <= from_lseq_mux;
                                        if (i_valid = '1') then
                                            c_Wcoef_next   <= c_Wcoef_r+1;
                                            outfifo_write <= '1';
                                        end if;
                                        wait_next <= '1';
                                    else
                                        if (i_valid = '1') then
                                            c_Wcoef_next   <= c_Wcoef_r+1;
                                            outfifo_write <= '1';
                                        end if;
                                    end if;
                                end if;
                            end if;
                        end if;

                    end if;

                end if;

            end if;


            if (fifo_read = '1')then
                c_Rcoef_next   <= c_Rcoef_r+1;
            end if;

            case checkpoint is
                when "0000" =>
                    if (c_hashw_r >= 19)  then
                        if (c_Wcoef_r < 14)  then
                            chk_failed_next <= '1';
                        else
                            checkpoint_next   <= "0001";
                        end if;
                    end if;

                when "0001" =>
                    if (c_hashw_r >= 39)  then
                        if (c_Wcoef_r < 30)  then
                            chk_failed_next <= '1';
                        else
                            checkpoint_next   <= "0010";
                        end if;
                    end if;

                    if (c_Rcoef_r > 13) then
                        bypass_empty <= '1';
                    else
                        bypass_empty <= '0';
                    end if;

                when "0010" =>
                    if (c_hashw_r >= 59)  then
                        if (c_Wcoef_r < 47)  then
                            chk_failed_next <= '1';
                        else
                            checkpoint_next   <= "0011";
                        end if;
                    end if;

                    if (c_Rcoef_r > 29) then
                        bypass_empty <= '1';
                    else
                        bypass_empty <= '0';
                    end if;

                when "0011" =>
                    if (c_hashw_r >= 78)  then
                        if (c_Wcoef_r < 62)  then
                            chk_failed_next <= '1';
                        else
                            checkpoint_next   <= "0100";
                        end if;
                    end if;

                    if (c_Rcoef_r > 46) then
                        bypass_empty <= '1';
                    else
                        bypass_empty <= '0';
                    end if;

                when "0100" =>
                    if (c_hashw_r >= 98)  then
                        if (c_Wcoef_r < 79)  then
                            chk_failed_next <= '1';
                        else
                            checkpoint_next   <= "0101";
                        end if;
                    end if;

                    if (c_Rcoef_r > 61) then
                        bypass_empty <= '1';
                    else
                        bypass_empty <= '0';
                    end if;

                when "0101" =>
                    if (c_hashw_r >= 118)  then
                        if (c_Wcoef_r < 94)  then
                            chk_failed_next <= '1';
                        else
                            checkpoint_next   <= "0110";
                        end if;
                    end if;

                    if (c_Rcoef_r > 78) then
                        bypass_empty <= '1';
                    else
                        bypass_empty <= '0';
                    end if;

                when "0110" =>
                    if (c_hashw_r >= 137)  then
                        if (c_Wcoef_r < 110)  then
                            chk_failed_next <= '1';
                        else
                            checkpoint_next   <= "0111";
                        end if;
                    end if;

                    if (c_Rcoef_r > 93) then
                        bypass_empty <= '1';
                    else
                        bypass_empty <= '0';
                    end if;

                when "0111" =>
                    if (c_hashw_r >= 157)  then
                        if (c_Wcoef_r < 126)  then
                            chk_failed_next <= '1';
                        else
                            if (PARAM_N = 1087) then
                                checkpoint_next   <= "1010";
                            else
                                checkpoint_next   <= "1000";
                            end if;
                        end if;
                    elsif (PARAM_N = 1087) and (hash_done_r = '1') then
                        chk_failed_next <= '1';
                    end if;

                    if (c_Rcoef_r > 109) then
                        bypass_empty <= '1';
                    else
                        bypass_empty <= '0';
                    end if;

                when "1000" =>
                    if (c_hashw_r >= 177)  then
                        if (c_Wcoef_r < 142)  then
                            chk_failed_next <= '1';
                        else
                            checkpoint_next   <= "1001";
                        end if;
                    end if;

                    if (c_Rcoef_r > 125) then
                        bypass_empty <= '1';
                    else
                        bypass_empty <= '0';
                    end if;


                when "1001" =>
                    if (c_hashw_r >= 196) then
                        if (c_Wcoef_r < 158)  then
                            chk_failed_next <= '1';
                        else
                            checkpoint_next   <= "1010";
                        end if;
                    elsif (hash_done_r = '1') then
                        chk_failed_next <= '1';
                    end if;

                    if (c_Rcoef_r > 141) then
                        bypass_empty <= '1';
                    else
                        bypass_empty <= '0';
                    end if;

                when others =>
                    if (PARAM_N = 1087) then
                        if (c_Rcoef_r > 125) then
                            done_next    <= '1';
                            bypass_empty <= '1';
                        else
                            bypass_empty <= '0';
                        end if;
                    else
                        if (c_Rcoef_r > 157) then
                            done_next    <= '1';
                            bypass_empty <= '1';
                        else
                            bypass_empty <= '0';
                        end if;
                    end if;

            end case;

        when S_WAIT =>
            bypass_empty <= '1';
            done         <= '1';

    end case;
end process;

--! DATAPATH
Param_data_size_1087: if PARAM_N = 1087 generate
NUM_HASH32_REQ <= NUM_HASH_REQ_BWC1_1087*8;
end generate;

Param_data_size_1499: if PARAM_N = 1499 generate
NUM_HASH32_REQ <= NUM_HASH_REQ_BWC1_1499*8;
end generate;

chk_failed <= chk_failed_r;
hashw_flag <= ((not done_r) and (not outfifo_full));
wcoef_flag <= ((not load_r) and hashw_flag and (not first_coef_r) and i_valid)or
    (load_r and hashw_flag and (not wait_r) and i_valid);

Piso: entity work.piso
    generic map(
        N => W,
        M => piso_width
        )
    port map(
        clk    => clk,
        rst    => rst,
        en     => en_piso,
        ld     => ld_piso,
        input  => infifo_dout,
        output => piso_out
    );

Reg_piso: entity work.regn(struct)
    generic map(
        N    => piso_width-1,
        init => ZERO(piso_width-2 downto 0)
        )
	port map(
        clk    => clk,
        rst    => rst,
        en     => en_piso,
        input  => piso_out(piso_width-2 downto 0),
        output => piso_r
        );

Var_shift : entity work.var_shiftr(vsr)
    generic map( N => 28 )
    port map(
        A => piso_r & piso_out(piso_width-1 downto 3),
        B => vs_rom_out,
        C => vs_out
    );

VS_ROM : entity work.vs_rom(arch)
    port map(
        addr   => count_r(3 downto 0),
        output => vs_rom_out
    );

Mod_1499_ROM : entity work.mod1499_rom
    generic map(
        ADDR_WIDTH => c,
        DATA_WIDTH => BETA
        )
    port map(
        addr   => vs_out(c-1 downto 0),
        output => mod_1499
        );

Mod_1087_ROM : entity work.mod1087_rom
    generic map(
        ADDR_WIDTH => c,
        DATA_WIDTH => BETA
        )
    port map(
        addr   => vs_out(c-1 downto 0),
        output => mod_1087
        );

from_mod_mux <= mod_1087 when (PARAM_N = 1087) else mod_1499;
--! c is valid (1) when we're under the threshold of 7495
lseq_7495 <= '1' when (vs_out(c-1 downto 0) <= "1110101000111") else '0';
--! c is valid (1) when we're under the threshold of 7609
lseq_7609 <= '1' when (vs_out(c-1 downto 0) <= "1110110111001") else '0';
from_lseq_mux <= lseq_7609 when (PARAM_N = 1087) else lseq_7495;

Reg_mod: entity work.regn(struct)
    generic map(
        N    => BETA,
        init => ZERO(BETA-1 downto 0)
        )
	port map(
        clk    => clk,
        rst    => rst,
        en     => '1',
        input  => from_mod_mux,
        output => from_mod_mux_r
        );

i_valid    <= (not dout_invalid) and from_lseq_mux_r;

Gen_behave_sim_regular_ram : if asym_ram = 0 generate

    RAM_i_reg : entity work.rams_tdp_rf_rf(syn)
        port map(
    		-- clock
    		clka    => clk,
            clkb    => clk,
    		-- port a
    		addra   => from_mod_mux,
    		dia     => ONES(ram_data_width-1 downto 0),
    		doa     => dout_invalid_stdv,
    		wea     => we_one,
            ena     => '1',
    		-- port b
    		addrb   => count_r,
    		dib     => ZERO(ram_data_width-1 downto 0),
    		dob     => open,
    		web     => we_zero,
            enb     => we_zero
        );
end generate;

Gen_behave_sim_asym_ram : if asym_ram = 1 generate
    RAM_i_asym : entity work.asym_ram_tdp_read_first(behavioral)
        generic map(
            WIDTHA => 1,
            SIZEA => 2048,
            ADDRWIDTHA => 11,
            WIDTHB => 32,
            SIZEB => 64,
            ADDRWIDTHB => 6
            )
        port map(
    		-- clock
    		clkA    => clk,
            clkB    => clk,
    		-- port a
    		addrA   => from_mod_mux,
    		diA     => ONES(ram_data_width-1 downto 0),
    		doA     => dout_invalid_stdv,
    		weA     => we_one,
            enA     => '1',
    		-- port b
    		addrB   => count_r(5 downto 0),
    		diB     => ZERO(31 downto 0),
    		doB     => open,
    		weB     => we_zero,
            enB     => we_zero
        );
end generate;

    dout_invalid <= dout_invalid_stdv(0);

    fifo_empty_valid <= '1' when (bypass_empty = '1') else outfifo_empty;

rstn <= not rst;
--! Note: we need to modify the controller to support stalls in fifo to reduce the size of fifo
Output_FIFO_FWFT: entity work.fifo(structure)
generic map(
    G_LOG2DEPTH    => 8,
    G_W            => BETA
)
port map (
    clk              => clk,
    rstn             => rstn,
    full             => outfifo_full,
    write            => outfifo_write,
    di_data          => from_mod_mux_r,
    empty            => outfifo_empty,
    read             => fifo_read,
    do_data          => fifo_dout,
    almost_full      => open,
    almost_empty     => open
);

end arch;
