-- =============================================================================
-- Copyright © 2017-2018 by Cryptographic Engineering Research Group (CERG),
-- ECE Department, George Mason University
-- Fairfax, VA, U.S.A.
-- Author: Farnoud Farahmand
-- =============================================================================
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_arith.all;

library work;
use work.sha2_pkg.all;
use work.sha3_pkg.all;

package ntru_pkg is
    --! NTRU SVES Parameter, Available Options 1087 and 1499
    --! Modifiable
    constant PARAM_N		 : integer 	:= 1499;



    --! Don't modify
    --! PQC Interface
    constant PW              : integer := 32;
    constant SW              : integer := 16;
    --! Headers
    constant MESSAGE         : std_logic_vector(2 downto 0) := "001";
    constant CIPHERTEXT      : std_logic_vector(2 downto 0) := "011";
    constant SYS_PARAM       : std_logic_vector(2 downto 0) := "100";
    constant PUBLIC_KEY      : std_logic_vector(2 downto 0) := "101";
    constant PRIVATE_KEY     : std_logic_vector(2 downto 0) := "110";
    --! Instructions
    constant INST_ENCRYPT    : std_logic_vector(3 downto 0) := "0000";
    constant INST_DECRYPT    : std_logic_vector(3 downto 0) := "0001";
    constant INST_LD_PRIVKEY : std_logic_vector(3 downto 0) := "0110";
    --! Status
    constant SUCCESS         : std_logic_vector(3 downto 0) := "1110";
    constant FAILURE         : std_logic_vector(3 downto 0) := "1111";

    --! ODCU-E
    constant do_704_1087    : integer := 17;
    constant do_704_1499    : integer := 24;
    constant do_last_partial_1087 : integer := 63;
    constant do_last_partial_1499 : integer := 27;
    --! CHECK_1
    constant CHECK1_MinMsgRep_WT_1087 : integer := 63;
    constant CHECK1_MinMsgRep_WT_1499 : integer := 79;
    constant Last_Pad_1087            : integer := 1;
    constant Last_Pad_1499            : integer := 27;
    --! TB
    constant maxMsgLen_32_1087 : integer 	:= 45;
    constant maxMsgLen_32_1499 : integer 	:= 62;
    --! Top Controller
    constant maxMsgLen_Byte_1087 : integer 	:= 178;
    constant maxMsgLen_Byte_1499 : integer 	:= 247;
    constant MbinSize_32_1087: integer 	:= 51;
    constant MbinSize_32_1499: integer 	:= 70;
    --! BGPM_MGF
	constant db_Byte_1087    : integer 	:= 24;
    constant db_Byte_1499    : integer 	:= 32;
    constant db_32_1087      : integer 	:= 6;
    constant db_32_1499      : integer 	:= 8;
    constant pkLen_Byte_1087 : integer 	:= 24;
    constant pkLen_Byte_1499 : integer 	:= 32;
    constant OID_Byte        : integer 	:= 3;
    constant Ci_Byte         : integer 	:= 4;
    constant Total_BPGM_Bytes_1087: integer := OID_Byte+db_Byte_1087+pkLen_Byte_1087;
    constant Total_BPGM_Bytes_1499: integer := OID_Byte+db_Byte_1499+pkLen_Byte_1499;

    constant Total_MGF_Bytes_1087: integer := (17*16)+Ci_Byte;
    constant Total_MGF_Bytes_1499: integer := (24*16)+Ci_Byte;

    constant MGF_32bits_FirstChunk_1087: integer := 64;
    constant MGF_Full_32bits_LastChunk_1087: integer := 4;
    constant MGF_32bits_LastChunk_1087: integer := 5;
    constant MGF_bits_LastChunk_1087: integer := 160;
    constant MGF_Partialbits_Last32Chunk_1087: integer := 32;
    constant MGF_LastBLK_PISO_Shift_1087: integer := 3;

    constant MGF_32bits_FirstChunk_1499: integer := 80;
    constant MGF_Full_32bits_LastChunk_1499: integer := 12;
    constant MGF_32bits_LastChunk_1499: integer := 15;
    constant MGF_bits_LastChunk_1499: integer := 472;
    constant MGF_Partialbits_Last32Chunk_1499: integer := 24;
    constant MGF_LastBLK_PISO_Shift_1499: integer := 1;

    constant NUM_HASH_REQ_BWC1_1087: integer  := 8;
    constant NUM_HASH_REQ_BWC1_1499: integer  := 10;
    constant NUM_HASH_REQ_BWC2_1087: integer  := 9;
    constant NUM_HASH_REQ_BWC2_1499: integer  := 11;

	constant OID_1087 : std_logic_vector(23 downto 0) := x"000505";
    constant OID_1499 : std_logic_vector(23 downto 0) := x"000605";

    constant CNT_SIZE : integer := 11;
    constant asym_ram : integer := 0;

    constant ZERO  : std_logic_vector(127 downto 0) := (others => '0');

    --! POLY_MULT
    constant ALPHA 				: integer := 11;
	constant BETA 				: integer := 11;
    constant DF_1087    : integer := 63;
    constant DF_1499    : integer := 79;

    constant PIPE_N : std_logic_vector(BETA-1 downto 0) := "10100010010"; --"00000000000"
    constant PIPE_STAGES : integer := 4;

	constant dm0 				: std_logic_vector(7 downto 0) := "00111111";	-- 01001111 dm0 = 79

    constant N                  : integer := PARAM_N;
    constant LM 				: integer := 16;
	constant ZEROs 				: std_logic_vector(ALPHA*PARAM_N-1 downto 0) := (others => '0');

    function has_rmndr(X : integer) return integer;
    function divceil(a  : natural; b : natural) return natural;
    function log2ceil(num : natural) return natural;
    function slv_to_string( v : std_logic_vector) return string;

end ntru_pkg;

-- functions description
package body ntru_pkg is
    function has_rmndr(X : integer) return integer is
    begin
        if(X > 0) then
            return 1;
        else
            return 0;
        end if;
    end function;

    function divceil (a : natural; b : natural ) return natural is
    begin
        if(a mod b) > 0 then
            return (a / b) + 1;
        else
            return (a / b);
        end if;
    end function;

    function log2ceil ( num : natural) return natural is
        variable i    : natural;
        variable bits : natural;
    begin
        i := 0;
        while(2**i < num) loop
            i := i + 1;
        end loop;
        return i;
    end function;

    function slv_to_string( v : std_logic_vector) return string is
        variable s : string (v'length-1 downto 0) := (others => NUL);
    begin
        for i in v'length - 1 downto 0 loop
            s(i) := std_logic'image(v(i))(2);
        end loop;
        return s;
    end function;
end package body ntru_pkg;

-- ==========================================================================
