-- =============================================================================
-- Copyright © 2017-2018 by Cryptographic Engineering Research Group (CERG),
-- ECE Department, George Mason University
-- Fairfax, VA, U.S.A.
-- Author: Farnoud Farahmand
-- =============================================================================
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.numeric_std.all;

library work;
use work.ntru_pkg.all;

-- Output Data Conversion Unit for Encryption
entity odcu_e is
generic
(
	iw : integer := 704;
	ow : integer := 32
);
port
(
    clk       : in std_logic;
    rst       : in std_logic;
	check_1   : in std_logic;

    di_valid  : in std_logic;
    di_ready  : out std_logic;
    di        : in std_logic_vector(iw-1 downto 0);

    do_valid  : out std_logic;
    do_ready  : in std_logic;
	do_data   : out std_logic_vector(ow-1 downto 0)
);
end odcu_e;

-- ==========================================================================
architecture arch of odcu_e is

    constant zeros  : std_logic_vector(20 downto 0) := (others => '0');
    signal piso_out : std_logic_vector(10 downto 0);
    signal header   : std_logic_vector(31 downto 0);

    type t_state is (S_RESET, S_SEND_DATA, S_SEND_STATUS, S_LD_PISO, S_SEND_HDR,
        S_IDLE);
    signal state, state_next : t_state;
    signal count_r, count_next              : std_logic_vector(5 downto 0);
    signal count2_r, count2_next            : std_logic_vector(5 downto 0);
    signal check_1_r, check_1_next          : std_logic;

    signal en_piso, ld_piso                 : std_logic;
    signal sel_do                           : std_logic;
    signal do_704, do_last_partial          : integer;

begin
    ODCUE_1087_PARAM_GEN: if PARAM_N = 1087 generate
        do_704          <= do_704_1087;
        do_last_partial <= do_last_partial_1087;
    end generate;

    ODCUE_1499_PARAM_GEN: if PARAM_N = 1499 generate
        do_704          <= do_704_1499;
        do_last_partial <= do_last_partial_1499;
    end generate;

    piso_1 : entity work.piso generic map ( M => 11, N => iw)
    port map ( clk => clk, rst => rst, en => en_piso, ld => ld_piso, input => di, output => piso_out);

    do_data <= header when sel_do = '1' else (zeros(20 downto 0) & piso_out);

    --! Controller
    p_fsm: process(clk)
    begin
        if rising_edge(clk) then
            if (rst = '1') then
                state       <= S_RESET;
            else
                state       <= state_next;
            end if;
            count_r         <= count_next;
            count2_r        <= count2_next;
			check_1_r       <= check_1_next;
        end if;
    end process;

    p_comb: process(state, di_valid, do_ready, count_r, count2_r, check_1,
        check_1_r)
    --check_1_valid
    begin
        --! Default values
        state_next       <= state;
        count_next       <= count_r;
        count2_next      <= count2_r;
		check_1_next     <= check_1_r;

        header           <= (others => '0');
        di_ready         <= '0';
        do_valid         <= '0';
        en_piso          <= '0';
        ld_piso          <= '0';
        sel_do           <= '0';

        case state is
            when S_RESET =>
                count2_next   <= (others => '0');
                check_1_next  <= '0';
                state_next    <= S_IDLE;

            when S_IDLE =>
                di_ready  <= '1';
                if (di_valid = '1') then
                    en_piso        <= '1';
                    ld_piso        <= '1';
                    count_next	   <= (others => '0');
                    count2_next    <= count2_r + 1;
                    state_next     <= S_SEND_HDR;
                end if;

            when S_SEND_HDR =>
                do_valid  <= '1';
                sel_do    <= '1';
                --! size is the number of 11 bits instead of 8 bits
                header <= CIPHERTEXT & "00011" & X"00" &
                    std_logic_vector(to_unsigned(PARAM_N, 16));
                if (do_ready = '1') then
                    state_next  <= S_SEND_DATA;
                end if;

            when S_SEND_DATA =>
                do_valid    <= '1';
                if (do_ready = '1') then
                    if (count2_r = do_704) then
                        if (count_r =  do_last_partial-1) then
                            count_next <= (others => '0');
                            state_next <= S_SEND_STATUS;
                        else
                            en_piso    <= '1';
                            count_next <= count_r + 1;
                        end if;
                    else
                        if (count_r = (iw/11)-1) then
                            count_next <= (others => '0');
                            state_next <= S_LD_PISO;
                        else
                            en_piso    <= '1';
                            count_next <= count_r + 1;
                        end if;
                    end if;
                end if;

            when S_LD_PISO =>
                di_ready  <= '1';
                if (di_valid = '1') then
                    en_piso        <= '1';
                    ld_piso        <= '1';
                    count_next	   <= (others => '0');
                    count2_next    <= count2_r + 1;
                    state_next     <= S_SEND_DATA;
                    if (count2_r = do_704-1) then
                        check_1_next <= check_1;
                    end if;
                end if;

            when S_SEND_STATUS =>
                do_valid  <= '1';
                sel_do    <= '1';
				if (check_1_r = '1') then
                	header <= SUCCESS & X"0000000";
				else
					header <= FAILURE & X"0000000";
				end if;
                if (do_ready = '1') then
                    state_next  <= S_RESET;
                end if;

        end case;
    end process;

end arch;

-- ==========================================================================
