-- =============================================================================
-- Copyright © 2017-2018 by Cryptographic Engineering Research Group (CERG),
-- ECE Department, George Mason University
-- Fairfax, VA, U.S.A.
-- Author: Kevin Briggs, Farnoud Farahmand
-- =============================================================================
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.NUMERIC_STD.ALL;

library work;
use work.ntru_pkg.all;

entity poly_mult_controller is
  generic (
      NUM : integer;        -- the number of coefficients
      NUM_SIZE : integer;   -- bit width of coefficients
      NUMPER : integer;     -- number of coefficients to load at a time
      ROTWIDTH : integer;
      OUT_WIDTH : integer;
      PIPE_STAGES : integer
  );
  Port (
         -- standard inputs
         clk      : in  std_logic;
         rst      : in  std_logic;
         -- external control inputs
         r_F_sel       : in std_logic;
         sdi_ready     : in std_logic;
         i_fifo_empty  : in std_logic;
         he_fifo_empty : in std_logic;
         ce_read       : in std_logic;
         Rfe_read      : in std_logic;
         param         : in std_logic;
         update_h      : in std_logic;
         decrypt       : in std_logic;
         start         : in std_logic;
         sdi_addr      : out std_logic_vector(8-1 downto 0);
         i_fifo_read   : out std_logic;
         he_fifo_read  : out std_logic;
         Rfe_valid     : out std_logic;
         ce_valid      : out std_logic;
         done          : out std_logic;
         -- to/from datapath
         c0       : out std_logic;
         sel_d    : out std_logic;
         sel_t    : out std_logic;
         sel_r    : out std_logic;
         ce_piso_en : out std_logic;
         ce_piso_ld : out std_logic;
         he_sipo_en : out std_logic;
         he_sipo_ld : out std_logic;
         Rfe_piso_en : out std_logic;
         Rfe_piso_ld : out std_logic;

         en_pipe   : out std_logic
       );
end poly_mult_controller;

architecture Behavioral of poly_mult_controller is

    constant num_out_cycles : integer := divceil((NUM_SIZE*NUM), (2*NUM_SIZE*OUT_WIDTH));

    type state_type is (S_INIT, S_LOAD_HE, S_SAVE_H, S_RESTORE_H, S_MULT, S_ADD_2FE, S_ADD_E, S_READ_OUT_BPG, S_DONE, S_READ_OUT_ADD, S_MULT_PIPE);
    signal state      : state_type;
    signal state_next : state_type;
    signal MULT_STEP : integer := 0;
    signal extra : integer := NUM - (NUM/NUMPER) * NUMPER;   --
    signal I_BOUND : integer := (NUM/NUMPER) - (1 - extra);  -- upper loop bound for loading coefficients

    signal i            : integer := 0;
    signal sdi_addr_int : integer := 0;
    signal DF : integer;

    -- Operation
    signal operation, operation_next : std_logic_vector(1 downto 0);
    constant LOAD_H    : std_logic_vector(1 downto 0) := "00";
    constant LOAD_E    : std_logic_vector(1 downto 0) := "01";

begin

    Param_1087: if PARAM_N = 1087 generate
        DF   <= DF_1087;
    end generate;
    Param_1499: if PARAM_N = 1499 generate
        DF   <= DF_1499;
    end generate;

    sdi_addr <= std_logic_vector(to_unsigned(sdi_addr_int, 8));

    seq : process(clk)
    begin
        if rising_edge(clk) then
            if(rst = '1') then
                state <= S_INIT;
                operation_next <= LOAD_H;
                ce_valid <= '0';
                sdi_addr_int <= 0;
                i <= 0;
            else
                state     <= state_next;
                operation <= operation_next;

                if(state = S_LOAD_HE) then
                    if(i < I_BOUND) and (he_fifo_empty = '0') then
                        i <= i + 1;
                    elsif(i = I_BOUND) and (he_fifo_empty = '0') then
                        i <= 0;
                    end if;
                elsif(state = S_SAVE_H) then
                    ce_valid <= '1';
                elsif(state = S_MULT) then
                    -- When r_F_sel = '0', we are reading coefficients of r from BPGM/MGF
                    -- When r_F_sel = '1', we are reading coefficients of F from RAM
                    if(r_F_sel = '0' and i_fifo_empty = '0') or (r_F_sel = '1' and sdi_ready = '1') then
                        if(i < PIPE_STAGES+(2*DF)-1) then
                            i <= i + 1;
                        else
                            i <= 0;
                        end if;
                        if(r_F_sel = '1') then
                            if(sdi_addr_int < 2*DF-1) then
                                sdi_addr_int <= sdi_addr_int + 1;
                            end if;
                        end if;
                    end if;
                elsif(state = S_MULT_PIPE) then
                    if(i < PIPE_STAGES+(2*DF)-1) then
                        i <= i + 1;
                    else
                        i <= 0;
                    end if;

                elsif(state = S_ADD_E) then
                    sdi_addr_int <= 0;
                    if(MULT_STEP < PIPE_STAGES) then
                        MULT_STEP <= MULT_STEP + 1;
                    else
                        MULT_STEP <= 0;
                    end if;
                elsif(state = S_READ_OUT_BPG) or (state = S_READ_OUT_ADD) then
                    if(Rfe_read = '1') then
                        if(i < num_out_cycles - 1) then
                            i <= i + 1;
                        else
                            i <= 0;
                        end if;
                    end if;
                end if;
            end if;
        end if;
    end process;


    comb : process(start, i, r_F_sel, sdi_ready, i_fifo_empty, he_fifo_empty, ce_read, Rfe_read, param, update_h, decrypt, start, state, MULT_STEP, operation)
    begin
        state_next     <= state;
        operation_next <= operation;

        -- external outputs
        i_fifo_read <= '0';
        he_fifo_read <= '0';
        Rfe_valid <= '0';
        done <= '0';
        -- to datapath
        c0       <= '0';
        sel_d    <= '0';
        sel_t    <= '0';
        sel_r    <= '0';
        ce_piso_en <= '0';
        ce_piso_ld <= '0';
        he_sipo_en <= '0';
        he_sipo_ld <= '0';
        Rfe_piso_en <= '0';
        Rfe_piso_ld <= '0';

        en_pipe     <= '0';

        case state is
            when S_INIT =>
                if(start = '1' and update_h = '1') then
                    operation_next <= LOAD_H;
                    state_next <= S_LOAD_HE;
                elsif(start = '1' and decrypt = '0') then
                    state_next <= S_MULT;
                elsif(start = '1' and decrypt = '1' and r_F_sel = '0') then
                    state_next <= S_MULT;
                elsif(start = '1' and decrypt = '1' and r_F_sel = '1') then
                    operation_next <= LOAD_E;
                    state_next <= S_SAVE_H;
                end if;

            when S_LOAD_HE =>
                -- The controller needs to know how many cycles to load data into the SIPO for.
                if(he_fifo_empty = '0') then
                    he_sipo_en <= '1';
                    he_fifo_read <= '1';
                    if(i = I_BOUND) then
                        if (operation = LOAD_H) then
                            done  <= '1'; --!test
                            state_next <= S_INIT;
                        else
                            state_next <= S_MULT;
                        end if;
                    end if;
                end if;
            -- load pubkey into top piso
            when S_SAVE_H  =>
                -- load pubkey into piso0 in either mode
                ce_piso_en <= '1';
                ce_piso_ld <= '1';
                operation_next <= LOAD_E;
                state_next     <= S_LOAD_HE;
            when S_RESTORE_H =>
                -- restore pubkey into sipo
                he_sipo_en <= '1';
                he_sipo_ld <= '1';
                -- load ciphertext to the ce piso
                ce_piso_en <= '1';
                ce_piso_ld <= '1';
                state_next <= S_READ_OUT_BPG;

            when S_MULT =>
                if(r_F_sel = '0' and i_fifo_empty = '0') or (r_F_sel = '1' and sdi_ready = '1') then
                    en_pipe     <= '1';
                    if(i >= PIPE_STAGES) then
                        Rfe_piso_en <= '1';
                        Rfe_piso_ld <= '1';
                    end if;
                    if(r_F_sel = '0') then
                        i_fifo_read <= '1';
                    end if;
                    if(i > PIPE_STAGES) then
                        sel_t <= '1';
                    end if;
                    if(i >= (PIPE_STAGES) + DF) then
                        c0 <= '1';
                    end if;

                    if (i = (2*DF)-1) then
                        if (PIPE_STAGES = 0) then
                            if(decrypt = '0') or ((decrypt = '1') and (r_F_sel = '0'))then
                                state_next <= S_DONE;
                            else
                                state_next <= S_ADD_2FE;
                            end if;
                        else
                            state_next <= S_MULT_PIPE;
                        end if;
                    end if;
                end if;

            when S_MULT_PIPE =>
                en_pipe     <= '1';
                Rfe_piso_en <= '1';
                Rfe_piso_ld <= '1';
                sel_t       <= '1';
                c0          <= '1';
                if (i =  (2*DF)+PIPE_STAGES-1) then
                    if(decrypt = '0') or ((decrypt = '1') and (r_F_sel = '0'))then
                        state_next <= S_DONE;
                    else
                        state_next <= S_ADD_2FE;
                    end if;
                end if;
                if (decrypt = '1') and (r_F_sel = '1') then
                    sel_r    <= '1';
                end if;

            when S_ADD_2FE =>
                sel_t <= '1';
                sel_d <= '1';
                Rfe_piso_en <= '1';
                Rfe_piso_ld <= '1';
                state_next <= S_ADD_E;

            when S_ADD_E =>
                Rfe_piso_en <= '1';
                Rfe_piso_ld <= '1';
                state_next <= S_DONE;
                sel_t <= '1';
                sel_r <= '1';

            when S_READ_OUT_BPG =>
                Rfe_valid <= '1';
                if(Rfe_read = '1') then
                    Rfe_piso_en <= '1';
                    if (r_F_sel = '1') or ((decrypt = '1')and(r_F_sel = '0'))then
                        ce_piso_en <= '1';
                    end if;
                    if(i = num_out_cycles - 1) then
                        if ((decrypt = '1')and(r_F_sel = '0')) then
                            state_next <= S_INIT;
                        else
                            state_next <= S_READ_OUT_ADD;
                        end if;
                        done <= '1';
                    end if;
                end if;

            when S_READ_OUT_ADD =>
                Rfe_valid <= '1';
                if(Rfe_read = '1') then
                    Rfe_piso_en <= '1';
                    if(i = num_out_cycles - 1) then
                        state_next <= S_INIT;
                    end if;
                end if;

            when S_DONE =>
                done <= '1';
                if(start = '1') then
                    if (r_F_sel = '1') then
                        state_next <= S_RESTORE_H;
                    else
                        state_next <= S_READ_OUT_BPG;
                    end if;
                end if;
        end case;
    end process;
end Behavioral;
