---=========================================================
--- Project      : Hardware Implementation of NTRU Cryptosystems
--- Author       : Kevin Briggs
---=========================================================
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_arith.all;

library work;
use work.sha2_pkg.all;
use work.sha3_pkg.all;
use work.ntru_pkg.all;

entity poly_mult_top is
generic (
    N: integer;
    ALPHA : integer;
    BETA  : integer;
    W     : integer;
    -- DF    : integer;
    PIPE_N : std_logic_vector;
    PIPE_STAGES : integer
);
port
(
    -- global inputs
	clk 	    	: in std_logic;
	rst_dp 	      : in std_logic;
    rst_ctrl      : in std_logic;

	r_F_sel       : in std_logic;
	-- sdi_addr      : out std_logic_vector(log2ceil(2*DF)-1 downto 0);
    sdi_addr      : out std_logic_vector(8-1 downto 0);
	sdi_ready     : in std_logic;
	i_fifo_empty  : in std_logic;
	i_fifo_read   : out std_logic;
	he_fifo_empty : in std_logic;
	he_fifo_read  : out std_logic;
	ce_valid      : out std_logic;
	ce_read       : in std_logic;
	Rfe_valid     : out std_logic;
	Rfe_read      : in std_logic;
	param         : in std_logic;
	update_h      : in std_logic;
	decrypt       : in std_logic;
	start         : in std_logic;
	done          : out std_logic;
    r_F           : in std_logic_vector(BETA-1 downto 0);
    h_e           : in std_logic_vector((ALPHA*(W/ALPHA))-1 downto 0);
    ce            : out std_logic_vector(2*W*ALPHA-1 downto 0);
    R_fe          : out std_logic_vector(2*W*ALPHA - 1 downto 0);
    cR            : in std_logic_vector(2*W*ALPHA-1 downto 0)

);
end poly_mult_top;

architecture arch of poly_mult_top is

	-- signals
	signal piso             : std_logic_vector(4*W*ALPHA-1 downto 0);
	signal piso_data        : std_logic_vector(4*W*ALPHA-1 downto 0);

	-- constants
	constant df_1499        : integer := 79;

	signal r_f_z_sel_sig : std_logic;
	signal mode_sel_sig  : std_logic;

	signal sel_t_sig      : std_logic;
	signal sel_d_sig      : std_logic;
	signal sel_r_sig      : std_logic;
	signal c0_sig         : std_logic;

	---- ce piso signals
	signal ce_piso_en_sig : std_logic;
	signal ce_piso_ld_sig : std_logic;

	-- he sipo signals
	signal he_sipo_en_sig : std_logic;
	signal he_sipo_ld_sig : std_logic;

	-- Rfe piso signals
	signal Rfe_piso_en_sig : std_logic;
	signal Rfe_piso_ld_sig : std_logic;

    signal en_pipe         : std_logic;

begin

    -- Poly Mult Controller
    poly_mult_controller: entity work.poly_mult_controller(Behavioral)
        generic map (
            NUM => N,
            NUM_SIZE => ALPHA,
            NUMPER => (W/ALPHA),
            ROTWIDTH => BETA,
            OUT_WIDTH => W,
            -- DF => DF,
            PIPE_STAGES => PIPE_STAGES
        )
    port map
    (
        clk             => clk,
        rst             => (rst_ctrl or rst_dp),
        r_F_sel         => r_F_sel,
        sdi_addr        => sdi_addr,
        sdi_ready       => sdi_ready,
        i_fifo_empty    => i_fifo_empty,
        i_fifo_read     => i_fifo_read,
        he_fifo_empty   => he_fifo_empty,
        he_fifo_read    => he_fifo_read,
        ce_valid        => ce_valid,
        ce_read         => ce_read,
        Rfe_valid       => Rfe_valid,
        Rfe_read        => Rfe_read,
        param           => param,
        update_h        => update_h,
        decrypt         => decrypt,
        start           => start,
        done            => done,
        -- to/from datapath
        c0              => c0_sig,
        sel_d           => sel_d_sig,
        sel_t           => sel_t_sig,
        sel_r           => sel_r_sig,
        ce_piso_en      => ce_piso_en_sig,
        ce_piso_ld      => ce_piso_ld_sig,
        he_sipo_en      => he_sipo_en_sig,
        he_sipo_ld      => he_sipo_ld_sig,
        Rfe_piso_en     => Rfe_piso_en_sig,
        Rfe_piso_ld     => Rfe_piso_ld_sig,
        en_pipe         => en_pipe
    );

	-- Poly Mult Full
	poly_mult_datapath :  entity work.poly_mult (arch_full)
	    generic map (
	        PIPE_N   => PIPE_N,
	        NUM      => N,
	        NUMWIDTH => ALPHA,
	        ROTWIDTH => BETA,
	        WORDSIZE => W
	    )
        port map
        (
            -- inputs
            clk          	=> clk,
            rst             => rst_dp,
            ce_piso_en  => ce_piso_en_sig,
            ce_piso_ld  => ce_piso_ld_sig,
            he_sipo_en  => he_sipo_en_sig,
            he_sipo_ld  => he_sipo_ld_sig,
            Rfe_piso_en => Rfe_piso_en_sig,
            Rfe_piso_ld => Rfe_piso_ld_sig,
            c0 		    => c0_sig,
            -- input data/control
            r_F         => r_F,
            h_e         => h_e,
            sel_d       => sel_d_sig,
            sel_t       => sel_t_sig,
            sel_r       => sel_r_sig,
            -- outputs
            ce 		      	=> ce,
            R_fe 		    => R_fe,
            cR          => cR,
            en_pipe         => en_pipe
        );
end arch;
