-- =====================================================================
-- Copyright © 2010-2012 by Cryptographic Engineering Research Group (CERG),
-- ECE Department, George Mason University
-- Fairfax, VA, U.S.A.
-- Author: CERG and Farnoud Farahmand
-- =====================================================================

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use work.sha3_pkg.all;
use work.sha2_pkg.all;

entity sha2_datapath_rs is
generic( n : integer :=HASH_SIZE_256/SHA2_WORDS_NUM; s :integer:=LOG_2_8; flag: integer:=HASH_BLOCKS_256; a :integer:=LOG_2_64; r:integer:=ROUNDS_64; cs: integer := LOG_2_1024 );
port(
	clk							: in std_logic;
	rst							: in std_logic;
	wr_state					: in std_logic;
	wr_result					: in std_logic;
	wr_data						: in std_logic;
	kw_wr						: in std_logic;
	wr_len						: in std_logic;
	sel							: in std_logic;
	sel2						: in std_logic;
	sel_gh						: in std_logic;
	sel_gh2						: in std_logic;
	ctr_ena						: in std_logic;
	z16							: out std_logic;
	zlast						: out std_logic;

	o8							: out std_logic;
	dst_write					: in std_logic;
	data,data_padded			: in std_logic_vector(n-1 downto 0);
	dataout						: out std_logic_vector(n-1 downto 0);
	wr_chctr					: in std_logic;
	last_block					: out std_logic;
	msg_done					: out std_logic;
	rst_flags					: in std_logic;
	 -- pad
	sel_pad_location, sel_din : out std_logic_vector(n/8-1 downto 0);
    en_lb, clr_lb : in std_logic; -- decrement last block word count ... rem1_reg
    comp_rem_e0, comp_lb_e0 : out std_logic;
    spos : in std_logic_vector(1 downto 0);
    --! NTRU control signals
    global_rst        : in std_logic;
    en_backup   : in std_logic;
    sel_init_state    : in std_logic;
    sel_input_state   : in std_logic
	);

end sha2_datapath_rs;

architecture sha2_datapath_rs of sha2_datapath_rs is
	type matrix 	is array (0 to STATE_REG_NUM-1) of std_logic_vector(n-1 downto 0);
	type matrix32 	is array (0 to STATE_REG_NUM-1) of std_logic_vector(31 downto 0);
	type matrix64 	is array (0 to STATE_REG_NUM-1) of std_logic_vector(63 downto 0);

	signal from_round			:matrix;
	signal to_round				:matrix;
	signal from_final_add		:matrix;
	signal from_mux				:matrix;
	signal result				:matrix;
	signal to_result			:matrix;
    --! NTRU signals
    signal from_init_mux		:matrix;
    signal from_state_backup	:matrix;
    signal from_instate_mux     :matrix;
    signal from_result_backup   :matrix;
    signal from_resinit_mux     :matrix;
    signal from_inresult_mux    :matrix;

	constant iv256					:matrix32 := ( SHA256_AINIT, SHA256_BINIT, SHA256_CINIT, SHA256_DINIT, SHA256_EINIT, SHA256_FINIT, SHA256_GINIT, SHA256_HINIT );
	constant iv512					:matrix64 := ( SHA512_AINIT, SHA512_BINIT, SHA512_CINIT, SHA512_DINIT, SHA512_EINIT, SHA512_FINIT, SHA512_GINIT, SHA512_HINIT );

	signal wwire				:std_logic_vector(n-1 downto 0);
	signal kwire				:std_logic_vector(n-1 downto 0);
	signal h_exception			:std_logic_vector(n-1 downto 0);
	signal rd_num				:std_logic_vector(a-1 downto 0);
	signal z16_reg				:std_logic;
	signal ena_reg				:std_logic;
	constant zero				:std_logic_vector(n-1 downto 0) := (others => '0');
	signal kwhwire				:std_logic_vector(n-1 downto 0);
	signal kwhreg				:std_logic_vector(n-1 downto 0);

    --! NTRU signals
    signal from_kwhinit_mux	    :std_logic_vector(n-1 downto 0);
    signal from_inkwh_mux		:std_logic_vector(n-1 downto 0);
    signal kwhreg_backup		:std_logic_vector(n-1 downto 0);

	signal chunk_ctr			:std_logic_vector(21 downto 0);
	constant BIT_SHF : integer := (n/32-1); -- use for shifting the msb of chunk ctr to the left by one

	signal out_ctr				:std_logic_vector(s downto 0);
	signal gh					:std_logic_vector(1 downto 0);
	signal last_block_out		:std_logic;

	signal remaining_bits : std_logic_vector(8+BIT_SHF downto 0);
    signal sel_pad_type_lookup, sel_din_lookup : std_logic_vector(n/8-1 downto 0 );

    type padsel_type is array (0 to 7) of std_logic_vector(8 downto 0);
	constant lookup_padsel : padsel_type := (
		"100000000", "110000000", "111000000", "111100000",
		"111110000", "111111000", "111111100", "111111110");
begin

	st256_gen: if n=HASH_SIZE_256/STATE_REG_NUM generate
		sr_gen: for i in 0 to STATE_REG_NUM-1 generate
		sr0	: entity work.regn(struct) generic map (n=>n, init =>iv256(i))
					port map (clk=>clk, en=>wr_state, rst=>global_rst, input=>from_instate_mux(i), output=>to_round(i));
		end generate;
	end generate;

    --! NTRU state backup
    st256_backup_gen: if n=HASH_SIZE_256/STATE_REG_NUM generate
        sr_backup_gen: for i in 0 to STATE_REG_NUM-1 generate
        sr0_backup : entity work.regn(struct) generic map (n=>n, init => X"00000000")
            port map (
                clk    =>clk,
                en     =>en_backup,
                rst    =>global_rst,
                input  =>to_round(i),
                output =>from_state_backup(i)
                );
        end generate;
    end generate;
    state_backup_mux_gen1: for i in 0 to STATE_REG_NUM-1 generate
		from_init_mux(i) <= from_state_backup(i) when sel_init_state='1' else iv256(i);
	end generate;
    state_backup_mux_gen2: for i in 0 to STATE_REG_NUM-1 generate
		from_instate_mux(i) <= from_init_mux(i) when sel_input_state='1' else from_mux(i);
	end generate;
    --! -----------------------------------------------------------

	st512_gen: if n=HASH_SIZE_512/STATE_REG_NUM generate
		sr_gen: for i in 0 to STATE_REG_NUM-1 generate
		sr0	: entity work.regn(struct) 	generic map (n=>n, init =>iv512(i))
					port map (clk=>clk, en=>wr_state, rst=>rst, input=>from_mux(i), output=>to_round(i));
		end generate;
	end generate;

	kwh_reg		: entity work.regn(struct) 	generic map (n=>n, init =>zero)
	port map (clk=>clk, en=>kw_wr, rst=>global_rst, input=>from_inkwh_mux, output=>kwhreg);

    --! NTRU kwh backup
    kwh_backup_reg : entity work.regn(struct) generic map (n=>n, init =>zero)
    port map (clk=>clk, en=>en_backup, rst=>global_rst, input=>kwhreg, output=>kwhreg_backup);

    from_kwhinit_mux <= kwhreg_backup when sel_init_state='1' else zero;
    from_inkwh_mux   <= from_kwhinit_mux when sel_input_state='1' else kwhwire;
    --! -----------------------------------------------------------

	gh <= (sel_gh, sel_gh2);

	with gh select
	h_exception <= 	to_round(6) when "00",
					to_round(6) when "01",
					to_round(7) when "10",
					result(6) when OTHERS;


	round: 	entity work.sha2_round_rs(basic)
			generic map (n=>n)
			port map (
			sel_gh=>sel_gh, kw=>kwhreg, kwire=>kwire, wwire=>wwire,	ain=>to_round(0), bin=>to_round(1),
			cin=>to_round(2), din=>to_round(3), ein=>to_round(4), fin=>to_round(5), gin=>to_round(6),
			hin=>h_exception, kwhwire=>kwhwire, aout=>from_round(0),bout=>from_round(1), cout=>from_round(2),
			dout=>from_round(3), eout=>from_round(4), fout=>from_round(5), gout=>from_round(6), hout=>from_round(7));

		from_final_add(0)	<= to_round(0) +  result(3);
		from_final_add(4)	<= to_round(4) +  result(7);

	mux0_gen: for i in 0 to STATE_REG_NUM-1 generate
		from_mux(i) <= from_final_add(i) when sel='1' else from_round(i);
	end generate;

		from_final_add(1) <= result(0);
		from_final_add(2) <= result(1);
		from_final_add(3) <= result(2);
		from_final_add(5) <= result(4);
		from_final_add(6) <= result(5);
		from_final_add(7) <= result(6);

		ena_reg <=wr_result or dst_write or sel_input_state;

	mux1_gen: for i in 0 to STATE_REG_NUM-2 generate
			 to_result(i) <= from_final_add(i) when wr_result='1' else result(i+1);
	end generate;

		to_result(7) <= from_final_add(7);

	hs256_gen: if n=HASH_SIZE_256/STATE_REG_NUM generate
		rr_gen: for i in 0 to STATE_REG_NUM-1 generate
		rr	: entity work.regn(struct) 		generic map (n=>n, init =>iv256(i))
							port map (clk=>clk, en=>ena_reg, rst=>global_rst, input=>from_inresult_mux(i), output=>result(i));
		end generate;
	end generate;

    --! NTRU result backup
    hs256_backup_gen: if n=HASH_SIZE_256/STATE_REG_NUM generate
        rr_backup_gen: for i in 0 to STATE_REG_NUM-1 generate
        rr_backup : entity work.regn(struct)
                generic map (n=>n, init =>X"00000000")
                port map(
                    clk=>clk,
                    en=>en_backup,
                    rst=>global_rst,
                    input=>result(i),
                    output=>from_result_backup(i)
                );
        end generate;
    end generate;
    res_backup_mux_gen1: for i in 0 to STATE_REG_NUM-1 generate
        from_resinit_mux(i) <= from_result_backup(i) when sel_init_state='1' else iv256(i);
    end generate;
    res_backup_mux_gen2: for i in 0 to STATE_REG_NUM-1 generate
        from_inresult_mux(i) <= from_resinit_mux(i) when sel_input_state='1' else to_result(i);
    end generate;
    --! -----------------------------------------------------------

	hs512_gen: if n=HASH_SIZE_512/STATE_REG_NUM generate
		rr_gen: for i in 0 to STATE_REG_NUM-1 generate
		rr	: entity work.regn(struct) 		generic map (n=>n, init =>iv512(i))
							port map (clk=>clk, en=>ena_reg, rst=>rst, input=>to_result(i), output=>result(i));
		end generate;
	end generate;

	dc			: entity work.sha2_msg_scheduler(mc_evoy)
				generic map (n=>n)
				port map (clk=>clk, sel=>sel2, wr_data=>wr_data, data=>data_padded, w=>wwire);

	rd_ctr 		:entity work.sha2_ctr(sha2_ctr) 	generic map (s=>a, r=>r-1, step=>1)
				port map (clk=>clk, reset=>rst, ena=>ctr_ena, ctr=>rd_num);



	bf			:entity work.sha2_flags_bl(sha2_flags_bl) 	generic map (n=>n, a=>a, r=>r, s=>s, flag=>flag)
							port map (clk=>clk, en=>ctr_ena, wr_len=>wr_len, rd_num=>rd_num, exam_block=>data,
							out_ctr=>out_ctr, rst=>rst, rst_flags=>rst, z16=>z16_reg,
							zlast=>zlast, o8=>o8, last_block_in=>data(n-1), last_block_out=>last_block_out );

	last_block <= last_block_out;

	o_ctr		:entity work.sha2_ctr(sha2_ctr) 	generic map (s=>s+1, r=>flag+1)
							port map (clk=>clk, reset=>rst, ena=>dst_write, ctr=>out_ctr);

	const		:entity work.sha2_cons(sha2_cons) generic map (n=>n, a=>a)
				port map (address=>rd_num, output=>kwire);

	---//////////////---
	-- block decounter
	decounter_gen : decountern
		generic map ( N => 22, sub => 1 )
		port map ( clk => clk, rst => '0', load => wr_len, en => wr_chctr, input => data(30+BIT_SHF downto 9+BIT_SHF), output => chunk_ctr);	 -- ignore MSB
	msg_done <= '1' when (chunk_ctr=0) else '0';
	------------------------------

	z16 <= z16_reg;
	dataout <=result(0);


	-- padding unit
	rem1_gen : decountern
		generic map ( N => 4, sub => 1 )
		port map ( clk => clk, rst => clr_lb, load => wr_len, en => en_lb, input => data(8+BIT_SHF downto 5+BIT_SHF), output => remaining_bits(8+BIT_SHF downto 5+BIT_SHF));
	rem2_reg : entity work.regn(struct) generic map ( N => 5+BIT_SHF, init => zero(4+BIT_SHF downto 0) )
		port map ( clk => clk, rst => clr_lb, en => wr_len, input => data(4+BIT_SHF downto 0), output => remaining_bits(4+BIT_SHF downto 0));

	pos256 : if n=HASH_SIZE_256/STATE_REG_NUM generate
        sel_din_lookup <= lookup_padsel(conv_integer(remaining_bits(4 downto 3)))(7 downto 4);
        sel_din <= (others => '1') when spos(1) = '1' else sel_din_lookup;

        sel_pad_type_lookup <= lookup_padsel(conv_integer(remaining_bits(4 downto 3)))(8 downto 5);
        sel_pad_location <= (others => '0') when spos(0) = '0' else sel_pad_type_lookup;
	end generate;
	pos512 : if n=HASH_SIZE_512/STATE_REG_NUM generate
        sel_din_lookup <= lookup_padsel(conv_integer(remaining_bits(5 downto 3)))(7 downto 0);
        sel_din <= (others => '1') when spos(1) = '1' else sel_din_lookup;

        sel_pad_type_lookup <= lookup_padsel(conv_integer(remaining_bits(5 downto 3)))(8 downto 1);
        sel_pad_location <= (others => '0') when spos(0) = '0' else sel_pad_type_lookup;
	end generate;


	comp_rem_e0 <= '1' when remaining_bits = 0 else '0';
	comp_lb_e0 <= '1' when remaining_bits(8+BIT_SHF downto 5+BIT_SHF) = 0 else '0';

end sha2_datapath_rs;
