---=========================================================
--- Project      : Hardware Implementation of NTRU Cryptosystems
--- Author       : Kevin Briggs
---=========================================================
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_arith.all;

library work;
use work.func_pkg.all;

-- serial input parallel output with parallel input
entity sipo_w_PI is
	generic (
		IS_WIDTH	: integer := 12;   -- input serial width
		IP_WIDTH    : integer := 40;   -- input parallel width
		OP_WIDTH	: integer := 40   -- output parallel width
	);
	port (
		clk 	: in std_logic;
		rst     : in std_logic;
		en     	: in std_logic;
		ld		: in std_logic;
		sin  	: in std_logic_vector(IS_WIDTH-1 downto 0);
		pin  	: in std_logic_vector(IP_WIDTH-1 downto 0);
		pout 	: out std_logic_vector(OP_WIDTH-1 downto 0)
	);
end sipo_w_PI;

architecture msbf of sipo_w_PI is
	-- constants
	-- a 'word' here is defined as a singular unit of input data
	-- this can be multiple coefficients at a time
	constant NUM_WORDS      : natural := OP_WIDTH / IS_WIDTH;
	constant RMNDR          : natural := OP_WIDTH - (NUM_WORDS * IS_WIDTH);
	constant FILL_WIDTH     : natural := (IS_WIDTH - RMNDR) mod IS_WIDTH;
	constant X              : integer := has_rmndr(RMNDR);

	-- signals
	constant ZEROS      : std_logic_vector(FILL_WIDTH-1 downto 0)     := (others => '0');
	constant ZEROS_FULL : std_logic_vector(IS_WIDTH-1 downto 0)       := (others => '0');

	-- arrays
	-- the 'word array' is the full array of all words input
	type word_array is array ( 0 to NUM_WORDS+X-1) of std_logic_vector(IS_WIDTH-1 downto 0);
	signal bsig, reg_in, reg_out : word_array;

begin

    reg_in(0) <= bsig(0) when (ld = '1') else sin;
    bsig(0) <= pin(IS_WIDTH - FILL_WIDTH - 1 downto 0) & ZEROS;
    gen_main_reg : entity work.regn_v2 (pos_edge) generic map (N => IS_WIDTH, init => ZEROS_FULL)
        port map(clk => clk, rst => rst, en => en, input => reg_in(0), output => reg_out(0));

    gen_pout_end1 : if(RMNDR = 0) generate
        pout(IS_WIDTH - 1 downto 0) <= reg_out(0);
    end generate;

    gen_pout_end2 : if (RMNDR > 0) generate
        pout(RMNDR - 1 downto 0) <= reg_out(0)(IS_WIDTH - 1 downto IS_WIDTH - RMNDR);
    end generate;

    -- main generate section
	gen_main : for i in 1 to NUM_WORDS generate

	    -- generate main reg muxes, excluding the first and last
        gen_main_mux: if (i > 0 and i < NUM_WORDS) generate
            reg_in(i) <= bsig(i) when (ld = '1') else reg_out(i-1);
            bsig(i) <= pin((i+1)*IS_WIDTH - FILL_WIDTH - 1 downto ((i) * IS_WIDTH - FILL_WIDTH));
            gen_main_reg : entity work.regn_v2 (pos_edge) generic map (N => IS_WIDTH, init => ZEROS_FULL)
                        port map(clk => clk, rst => rst, en => en, input => reg_in(i), output => reg_out(i));
            pout(((i+1)*IS_WIDTH - 1 - FILL_WIDTH) downto ((i)*IS_WIDTH - FILL_WIDTH)) <= reg_out(i);
        end generate;

        gen_last_mux : if(RMNDR > 0 and i = NUM_WORDS) generate
            reg_in(i) <= bsig(i) when (ld = '1') else reg_out(i-1);
            bsig(i) <= pin((i+1)*IS_WIDTH - FILL_WIDTH - 1 downto ((i) * IS_WIDTH - FILL_WIDTH));
            gen_main_reg : entity work.regn_v2 (pos_edge) generic map (N => IS_WIDTH, init => ZEROS_FULL)
                        port map(clk => clk, rst => rst, en => en, input => reg_in(i), output => reg_out(i));
            pout(((i+1)*IS_WIDTH - FILL_WIDTH - 1) downto ((i)*IS_WIDTH - FILL_WIDTH)) <= reg_out(i);
        end generate;
	end generate;
end msbf;
-- =============================================================================================================
architecture lsbf of sipo_w_PI is
    -- constants
	-- a 'word' here is defined as a singular unit of input data
	-- this can be multiple coefficients at a time
	constant NUM_WORDS      : natural := divceil(OP_WIDTH, IS_WIDTH);
	constant NUM_FULL       : natural := OP_WIDTH / IS_WIDTH;
	constant RMNDR          : natural := OP_WIDTH - (NUM_FULL * IS_WIDTH);
	constant FILL_WIDTH     : natural := (IS_WIDTH - RMNDR) mod IS_WIDTH;
	constant X              : integer := has_rmndr(RMNDR);

	-- signals
	constant ZEROS      : std_logic_vector(FILL_WIDTH-1 downto 0)     := (others => '0');
	constant ZEROS_FULL : std_logic_vector(IS_WIDTH-1 downto 0)       := (others => '0');

	-- arrays
	-- the 'word array' is the full array of all words input
	type word_array is array ( 0 to NUM_WORDS-1) of std_logic_vector(IS_WIDTH-1 downto 0);
	signal bsig, reg_in, reg_out : word_array;
begin
    -- main generate section
	gen_main : for i in NUM_WORDS - 1 downto 0 generate

	    -------- bsig --------
	    gen_bsig : if(i < NUM_WORDS - 1) generate
	        bsig(i) <= pin((i+1)*IS_WIDTH - 1 downto ((i) * IS_WIDTH));

	        pout(((i+1)*IS_WIDTH - 1) downto ((i)*IS_WIDTH)) <= reg_out(i);
	    end generate;

	    gen_bsig_end : if(i = NUM_WORDS - 1) and (RMNDR > 0) generate
	        bsig(NUM_WORDS - 1) <= ZEROS & pin(OP_WIDTH - 1 downto OP_WIDTH - (IS_WIDTH - FILL_WIDTH));
	        pout(OP_WIDTH - 1 downto OP_WIDTH-RMNDR) <= reg_out(i)(RMNDR - 1 downto 0);
	    end generate;

	    ---------- reg_in --------
	    gen_reg_in_end : if (i = NUM_WORDS - 1) generate
	        reg_in(i) <= bsig(NUM_WORDS-1) when (ld = '1') else sin;
	    end generate;

	    gen_reg_in : if(i < NUM_WORDS - 1) generate
	        reg_in(i) <= bsig(i) when (ld = '1') else reg_out(i+1);
	    end generate;

	    -------- registers --------
	    gen_main_reg : entity work.regn_v2 (pos_edge) generic map (N => IS_WIDTH, init => ZEROS_FULL)
            port map(clk => clk, rst => rst, en => en, input => reg_in(i), output => reg_out(i));
	end generate;
end lsbf;
